
### DATA CLEANING TO PREPARE DATASET ###

import pandas as pd
import numpy as np
from zlib import crc32
from scipy.ndimage.interpolation import shift
from sklearn.model_selection import StratifiedShuffleSplit, train_test_split
from sklearn.impute import SimpleImputer, KNNImputer
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from sklearn.preprocessing import OrdinalEncoder, OneHotEncoder, LabelEncoder
from sklearn.compose import ColumnTransformer
from stpstone.quantitative_methods.eda import ExploratoryDataAnalysis


class DataCleaning:

    def test_set_check_hash(self, identifier, test_ratio):
        '''
        REFERENCES: HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW, 
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S., 
            978-1-492-03264-9.
        DOCSTRING: SET HASH FROM THE ROW IDENTIFIER TO REMAIN A TEST SAMPLE STABLE EVEN THOUGH ITS
            AN UPDATE IN THE CONSULTED DATABASE FETCHED TO MEMORY
        INPUTS: IDENTIFIER, TESTE RATIO
        OUTPUTS: IDENTIFIER HASH
        '''
        return crc32(np.int64(identifier)) & 0xffffffff < test_ratio * 2**32

    def split_train_test(self, df_data, test_ratio=0.2, random_seed=42, stratify_col=None):
        '''
        REFERENCES: HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW, 
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S., 
            978-1-492-03264-9.
        DOCSTRING: CREATE A DATASET RANDOMLY WITHOUT DISTINGUISH OF ID HASH, SO WHEN THE DATASET 
            IS REFRESHED IT WOULD BE A CHANGE IN THE TESTING AND TRAINING SETS
        INPUTS: DATAFRAME DATA, TEST RATIO AND RANDOM SEED
        OUTPUTS: TUPLE OF DATAFRAMES WITH TRAINING SET AND TEST SET
        '''
        return train_test_split(df_data, test_size=test_ratio, random_state=random_seed, 
                                stratify=stratify_col)

    def split_stratified_train_test(self, df_data, col_name, n_splits=1, test_size=0.2,
                                    random_state_seed=42):
        '''
        REFERENCES: HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW, 
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S., 
            978-1-492-03264-9.
        DOCSTRING: STRATIFIED TECHNIQUE TO CREATE SAMPLES FROM THE ORIGINAL DATASET BY COL NAME, 
            AIMING TO NOT BE CHANGED BY A NEW FLOW OF DATA
        INPUTS: ORIGINAL DATAFRAME, ID COLUMN (STR), NUMBER OF SPLITS (1 AS DEFAULT), 
            TEST SIZE (0.2 AS DEFAULT) AND RANDOM STATE SEED (42 AS DEFAULT)
        OUTPUTS: TRAIN AND TEST SET DATAFRAMES
        '''
        # creating stratified categories
        split_class = StratifiedShuffleSplit(n_splits=n_splits, test_size=test_size,
                                             random_state=random_state_seed)
        for train_index, test_index in split_class.split(df_data, df_data[col_name]):
            df_strat_train_set = df_data.loc[train_index]
            df_strat_test_set = df_data.loc[test_index]
        return df_strat_train_set, df_strat_test_set
    
    def split_train_test_by_id(self, df_data, test_ratio, id_column):
        '''
        REFERENCES: HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW, 
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S., 
            978-1-492-03264-9.
        DOCSTRING: TEST SAMPLE OF DATA, RANDOMLY CHOSEN, TRYING TO PRESERVE ITS FEATURES 
            TO PERFORM STATISTICAL TESTS IN ORDER TO EXPLAIN BEAHVIOURS OF THE POPULATION, WHEREAS 
            THE TRAIN PORTION IS REMAINIG DATA PROVIDED TO ENHANCE THE CONCLUSIONS OF THE MODEL 
            PROPOSED. SPLITTING BY ID USES INSTANCE'S IDENTIFIERS
        INPUTS: DF DATA, TEST RATIO AND ID_COLUMN (WHEN ITS NOT AVAILABLE IN THE DATA FRAME THERE 
            ARE TWO OPTIONS: USING THE RESET_INDEX() METHOD IN PANDAS OR STABLE FEATURES AS 
            FOREIGNER KEYS)
        OUTPUTS: TUPLE OF DATAFRAMES WITH TRAINING SET AND TEST SET
        '''
        # defining ids from the data sample
        ids = df_data[id_column]
        # defining original hashes from the data set
        in_test_set = ids.apply(
            lambda id_: self.test_set_check_hash(id_, test_ratio))
        # returning sample test and training portion
        return df_data.loc[~in_test_set], df_data.loc[in_test_set]

    def create_category_stratified_train_test_set(self, df_data, id_column_original,
                                                  id_column_category, list_bins, list_labels):
        '''
        DOCSTRING: CREATE A CATEGORY TO BE USED AS A PROXY FOR STRATIFIED DATASET PURPOSES
        INPUTS: DATAFRAME, ID ORIGINAL COLUMN, ID CATEGORY COLUMN, LIST OF BINS AND LIST OF LABELS
        OUTPUTS: DATAFRAME WITH CATEGORY COLUMN
        '''
        # create a category, to be used as a proxy for stratification
        df_data[id_column_category] = pd.cut(df_data[id_column_original],
                                             bins=list_bins, labels=list_labels)
        return df_data

    def remove_category_stratified_train_test_set(self, df_train_set, df_test_set,
                                                  id_column_category):
        '''
        DOCSTRING: REMOVE CATEGORICAL COLUMN FROM TRAINING AND TEST DATASET
        INPUTS: DATAFRAME TRAINING SET, DATAFRAME TEST SET, AND ID OF CATEGORICAL COLUMN
        OUTPUTS: TUPLE WITH DATAFRAME TRAINING SET AND DATAFRAME TEST SET
        '''
        for set_ in (df_train_set, df_test_set):
            set_.drop(id_column_category, axis=1, inplace=True)
        return df_train_set, df_test_set

    def dataframe_id_column_prportions(self, df_data, id_column):
        '''
        REFERENCES: HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW, 
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S., 
            978-1-492-03264-9.
        DOCSTRING: DATAFRAME PROPORTIONS BY UNIQUE VALUES IN COLUMN ID
        INPUTS: DF_DATA AND ID COLUMN
        OUTPUTS: A GROUP BY OF PROPORTIONS, REGARDING TO THE TOTAL INSTANCES, EVALUATED OVER THE
            DATAFRAME COLUMN OF INTEREST
        '''
        return df_data[id_column].value_counts() / len(df_data)

    def compare_stratified_random_samples_propotions(self, df_data_original, df_data_random_set,
                                                     df_data_stratified_set, id_column):
        '''
        REFERENCES: HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW, 
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S., 
            978-1-492-03264-9.
        DOCSTRING: COMPARE STRATIFIED AND RANDOM SAMPLES PORPOTIONS REGARDING ORIGINAL DATASET
        INPUTS: DF_DATA, TEST RATIO, ID COLUMN, RANDOM SEED(42 AS DEFAULT), N SPLITS (1 AS DEFAULT)
        OUTPUTS: DATAFRAME WITH PROPORTIONS FOR EACH SAMPLING METHOD (STRATIFIED AND RANDOM) 
            REGARDING THE ORIGINAL DATA BASE
        '''
        # comparing porpotions for each sample
        df_compare_props = pd.DataFrame({
            'Overall': self.dataframe_id_column_prportions(df_data_original, id_column),
            'Stratified': self.dataframe_id_column_prportions(df_data_stratified_set,
                                                              id_column),
            'Random': self.dataframe_id_column_prportions(df_data_random_set, id_column)
        }).sort_index()
        df_compare_props['Rand. %error'] = 100 * df_compare_props[
            'Random'] / df_compare_props["Overall"] - 100
        df_compare_props['Strat. %error'] = 100 * df_compare_props[
            'Stratified'] / df_compare_props['Overall'] - 100
        # returning comparison dataframe of both sampling methods
        return df_compare_props

    def replace_nan_values(self, array_data, strategy=None, missing_values=np.nan, n_neighbors=None):
        '''
        DOCSTRING: REPLACE NAN WITH VALUES WITH DESIRED STRATEGY (MEAN, MEDIAN OR MOST FREQUENT), 
            AS 0 PLACEMENT, MEDIAN, OR MEAN
        INPUTS: DATAFRAME OF INTEREST, AND STRATEGY
        OUTPUTS: DICTIONARY WITH STRATEGY, ARRAY REPLACERS, SAMPLE INCOMPLETE ROWS BEFORE CHANGES, 
            SAMPLE INCOMPLETE ROWS AFTER CHANGES, DATAFRAME BEFORE ADJUSTMENTS AND DATAFRAME AFTER 
            ADJUSTMENTS
        '''
        # creating a copy of the original data just with numbers, since text will trigger an error
        #   when fitting a strategy for replacing nan values
        array_data_copy = array_data.copy()
        # creating imputer with the desired strategy
        if n_neighbors == None:
            imputer = SimpleImputer(strategy=strategy, missing_values=missing_values)
        else:
            imputer = KNNImputer(n_neighbors=n_neighbors, missing_values=np.nan)
        # fitting data
        imputer.fit(array_data_copy)
        # tansforming the output dataset
        array_data_copy = imputer.transform(array_data_copy)
        # retuning dict
        return {
            'strategy': imputer.strategy,
            'array_replacers': imputer.statistics_,
            'array_before_adustments': array_data,
            'array_after_adjustments': array_data_copy
        }

    def convert_categories_from_strings_to_array(self, array_data, list_idx_target_cols,
                                                 encoder_strategy='one_hot_encoder'):
        '''
        DOCSTRING: CONVERT CATEGORIES FROM STRING TO NUMBERS
        INPUTS: ARRAY OF DATA TO BE ENCODED - STRATEGIES HANDLED: ONE_HOT_ENCODER (UNIQUE 
            IDENTIFICATION ARRAY WITH N X N DIMENSION, BEING N THE NUMBER OF DIFERENT CATEGORIES), 
            ORDINAL_ENCODER (ORDINAL NUMBERS FROM 0 TO N_CLASS-1), 
            LABEL_ENCODER (ORDINAL NUMBERS FROM 0 TO N_CLASS-1)
        OUTPUTS: DICTIONARY WITH KEYS ARRAY LABELS, ARRAY DATA CATEGORIZED IN NUMBERS AND 
            ARRAY DATA CATEGORIZED IN STRINGS
        '''
        # checking wheter or not the array is a one hot encoding, when the category has one success
        #   category and a failure one
        if encoder_strategy == 'one_hot_encoder':
            #   column transformer
            ct = ColumnTransformer(transformers=[
                ('encoders', OneHotEncoder(), list_idx_target_cols)
            ], remainder='passthrough')
        elif encoder_strategy == 'ordinal_encoder':
            #   column transformer
            ct = ColumnTransformer(transformers=[
                ('encoders', OrdinalEncoder(), list_idx_target_cols)
            ], remainder='passthrough')
        elif encoder_strategy == 'label_encoder':
            #   column transformer - the label encoder is usually used for logistic regressions in 
            #       order to configure the dependent variable as either 0 or 1
            ct = LabelEncoder()
        # paramether bl_one_hot_encoding ought be a boolean, in case another value is given
        #   return an error to the user and drop out the script
        else:
            raise Exception(
                'Paramether bl_one_hot_encoding ought be a boolean.')
        # turn string categories to numbers - for label transform the array data must be 1-dimension
        array_categoric_numbers = np.array(ct.fit_transform(array_data))
        # return dictionay with categories and labels
        return {
            'array_data_categorized_numbers': array_categoric_numbers,
            'array_data_categorized_strings': array_data
        }

    def feature_scaling(self, array_data, type_scaler='normalization', tup_feature_range=(0,1)):
        '''
        REFERENCES:  “HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW, 
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S., 
            978-1-492-03264-9.”, 
            https://stackoverflow.com/questions/40758562/can-anyone-explain-me-standardscaler
        DOCSTRING: FEATURE SCALING NORMALIZATION (MIN-MAX SCALING, WITH A 0 TO 1 RANGE) OR 
            STANDARDISATION (Z SCORE, WITH A -3 TO 3 RANGE, LESS INFLUENCED BY OUTLIERS)-WISE
        INPUTS: ARRAY DATA AND TYPE OF FEATURE SCALER
        OUTPUTS: DICTIONARY WITH DATA MAX, DATA MIN, SCALE, N SAMPLES, ARRAY WITH ORIGINAL DATA AND 
            ARRAY WITH SCALED DATA
        '''
        # checking wheter the array is unidimensional and reshaping it
        array_data = ExploratoryDataAnalysis().reshape_1d_arrays(array_data)
        # defining feature scaling strategy
        if type_scaler == 'normalization':
            scaler = MinMaxScaler(feature_range=tup_feature_range)
        elif type_scaler == 'standardisation':
            scaler = StandardScaler()
        else:
            raise Exception('Strategy {} is not defined'.format(type_scaler))
        # fitting to the model
        scaler.fit(array_data)
        # transfom input array data
        array_data_transformed = scaler.transform(array_data)
        # return dictionary with relevant information
        return {
            'scaler': scaler,
            'scale': scaler.scale_,
            'n_samples_seen': scaler.n_samples_seen_,
            'array_original_data': array_data,
            'array_scaled_data': array_data_transformed
        }
        
    def remove_noise_from_data(self, data_test, data_train):
        '''
        REFERENCES: (MULTIOUTPUT CLASSIFICATION) https://colab.research.google.com/github/ageron/handson-ml2/blob/master/03_classification.ipynb#scrollTo=utQpplj4fGwa
        DOCSTRING: CLEAR NOISE AIMMING TO HELP ESTIMATORS TO FIT A DATASET AND PREDICT A PIXEL 
            TO A TARGET LABEL
        INPUTS: DATA TEST AND DATA TRAIN
        OUTPUTS: TUPLE WITH DATA TEST AND TRAINING, ORIGINAL AND ENHANCED
        '''
        # enhancing data pixels
        noise = np.random.randint(0, 100, (len(data_train), 784))
        data_train_original = data_train + noise
        noise = np.random.randint(0, 100, (len(data_test), 784))
        data_test_original = data_test + noise
        # returning orignal data test, training and each one enhanced
        return {
            'data_test_original': data_test_original,
            'data_test_enhanced': data_test,
            'data_training_original': data_train_original,
            'data_training_enhanced': data_train
        }

    def shift_image(self, image, dx, dy):
        '''
        REFERENCES: (DATA AUGMENTATION) https://colab.research.google.com/github/ageron/handson-ml2/blob/master/03_classification.ipynb#scrollTo=vkIfY1tAfGwf
        DOCSTRING: SHIFT IMAGE POSITION FOR DATA AUGMENTATION, OR TRAINING SET EXPANSION, IN ORDER 
            TO ENHANCE THE MODEL ESTIMATOR PERFORMANCE
        INPUTS: IMAGE, DX AND DY MOVEMENTS
        OUTPUTS: IMAGE
        '''
        image = image.reshape((28, 28))
        shifted_image = shift(image, [dy, dx], cval=0, mode="constant")
        return shifted_image.reshape([-1])



### API ANBIMA- MÉTODO POST PARA REALIZAÇÃO DE REQUISIÇÕES ###

import ast
import requests
import json
from stpstone.handling_data.str import StrHandler
from stpstone.handling_data.json import JsonFiles
from stpstone.setting._global_slots import YAML_ANBIMA
from stpstone.cals.handling_dates import DatasBR


class AnbimaDev:
    '''
    REFENCE PRODUTCS AND INDEXES: REFERENCES: https://developers.anbima.com.br/api-portal/pt-br/content/swagger-preços-e-índices#/
    REFERENCES FUNDS: https://developers.anbima.com.br/api-portal/pt-br/content/swagger-fundos#/Fundos/findByCodigoFundoUsingGET
    '''

    def access_token(self, client_id, client_secret):
        '''
        REFERENCES: https://developers.anbima.com.br/api-portal/pt-br
        DOCSTRING: TOKEN PARA CONSULTA DA API DA ANBIMA
        INPUTS: CLIET_ID E CLIENT_SECRET
        OUTPUTS: JSON
        '''
        base64_credentials = StrHandler().base64_encode(client_id, client_secret)
        url = 'https://api.anbima.com.br/oauth/access-token'
        headers = {
            'Content-Type': 'application/json',
            'Authorization': base64_credentials
        }
        payload = {
            'grant_type': 'client_credentials'
        }
        response = requests.request(
            method="POST", url=url, headers=headers, data=json.dumps(payload))
        jsonify_message = ast.literal_eval(StrHandler().find_between(
            str(response.text.encode('utf8')), "b'", "'"))
        return JsonFiles().send_json(jsonify_message)

    def response_requests(self, url, headers, payload):
        '''
        DOCSTRING: RESPOSTA PADRÃO PARA REQUISIÇÕES NO SWAGGER DO ANBIMA DEV
        INPUTS: URL, HEADERS E PAYLOAD
        OUTPUTS: DICT COM CÓDIGO DE RESPOSTA, MENSAGEM E URL
        '''
        response = requests.request(
            method="POST", url=url, headers=headers, params=payload)
        try:
            jsonify_message = ast.literal_eval(StrHandler().find_between(
                str(response.text.encode('utf8')), "b'", "'"))
            dict_message = {
                'code': response,
                'message': jsonify_message,
                'url': response.url
            }
        except:
            str_message = StrHandler().find_between(
                str(response.text.encode('utf8')), "b'", "'")
            dict_message = {
                'code': response,
                'message': str_message,
                'url': response.url
            }
        return dict_message

    def payload_consulta(self, combinacao_keys, data_interesse=DatasBR().sub_working_days(
            DatasBR().curr_date(), 1), page_int=0, size_int=1000,
            data_inicio=DatasBR().sub_working_days(DatasBR().curr_date(), 2),
            data_fim=DatasBR().sub_working_days(DatasBR().curr_date(), 1)):
        '''
        REFERENCES: https://developers.anbima.com.br/api-portal/pt-br/content/swagger-preços-e-índices#/
        DOCSTRING: CRIAÇÃO DE PARÂMETROS PARA SERVIREM COMO QUERIES PARA DIRECIONAR CONSULTA DO
            SERVIÇO
        INPUTS: COMBINAÇÃO KEYS (DATA, MES-ANO, PAGE-SIZE), DATA INTERESSE (DEFAULT D-1 ÚTIL), 
            PAGE INTEGER (DEFAULT 0), SIZE INTEGER (DEFAULT 1000)
        OUTPUTS: JSON
        '''
        if str(type(data_interesse)) == "<class 'datetime.date'>":
            data_interesse = data_interesse.strftime('%Y-%m-%d')
        if combinacao_keys == 'data':
            return {
                'data': data_interesse
            }
        elif combinacao_keys == 'mes-ano':
            return {
                'mes': data_interesse.month,
                'ano': data_interesse.year
            }
        elif combinacao_keys == 'page-size':
            return {
                'page': page_int,
                'size': size_int
            }
        elif combinacao_keys == 'data-inicio-fim':
            return {
                'data-inicio': data_inicio,
                'data-fim': data_fim
            }
        else:
            raise Exception(
                'Nenhuma combinação de keys para composição de query válida')

    def app_response(self, client_id, access_token, endpoint_correte, natureza_ativo,
                     app_interesse, payload_query):
        '''
        REFERENCES: https://developers.anbima.com.br/api-portal/pt-br/content/swagger-preços-e-índices#/Títulos%20Públicos/getUsingGET_34
        DOCSTRING: PROJEÇÕES PARA IGPM E IPCA
        INPUTS: CLIENT ID, ACCESS_TOKEN (RESULTADO DE UMA CONEXÃO NO OAUTH2), ENDPOINT CORRENTE
            (SANDBOX OU PRODUCAO), NATUREZA DO ATIVO (debentures, indices, indices+,
            titulos-publicos, cri-cra, fundos), APP DE INTERESSE URL (CONFERIR OS APPS DISPONÍVEIS
            NA REFERÊNCIA), PAYLOAD QUERY, APPS DE INTERESSE DA ANBIMA (POR PADRÃO UM 
            DICIONÁRIO COM TODOS OS APPS DISPONIBILIZADOS EM 22/09/2020), E ENPOINTS 
            (POR PADRÃO UM DICIONÁRIO
            COM TODOS OS ENDPOINTS DISPONÍVEIS EM 22/09/2020)
        TIPS: PARA SABER MAIS SOBRE AS KEYS DE APPS ANBIMA DEV E ENDPOINTS ANBIMA DEV POR FAVOR
            USAR O MÉTODO KEYS() E USAR O PAYLOAD CONSULTA (FUNÇÃO DA CLASSE) PARA COLHER O 
            PAYLOAD DE INTERESSE
        OUTPUTS: JSON
        '''
        url = YAML_ANBIMA['hosts_anbima_dev_fundos'][endpoint_correte] + \
            YAML_ANBIMA['apps_anbima_dev'][natureza_ativo][app_interesse]
        headers = {
            'accept': 'application/json',
            'client_id': client_id,
            'access_token': access_token
        }
        payload = payload_query
        return AnbimaDev().response_requests(url, headers, payload)


# client_id = CREDENTIALS_ANBIMA['client_id']
# password_id = CREDENTIALS_ANBIMA['client_secret']
# access_token = AnbimaDev().access_token(client_id, password_id)['access_token']
# print(access_token)
# print(AnbimaDev().app_response(client_id, access_token,
#                                'sandbox', 'titulos-publicos', 'vna',
#                                AnbimaDev().payload_consulta('data')))




### LIBRARY TO COLLECT MACROECONOMIC INFORMATIONS FROM USA

import pandas as pd
from requests import request
from bs4 import BeautifulSoup
from stpstone.settings._global_slots import YAML_USA_MACRO
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.object import HandlingObjects
from stpstone.handling_data.html import HtmlHndler
from stpstone.handling_data.dicts import HandlingDicts
from stpstone.loggs.db_logs import DBLogs
from stpstone.handling_data.str import StrHandler


class USAMacro:

    def fred_request(self, api_key, data_source, data_source_id, date_inf=None, date_sup=None,
                     url_root='https://api.stlouisfed.org/fred/{}', file_type='json',
                     date_format=None):
        '''
        REFERENCES: https://fred.stlouisfed.org/docs/api/fred/
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # adjusting variables types
        if date_format == 'dd/mm/yyyy':
            date_inf = DatesBR().str_date_to_datetime(date_inf)
            date_sup = DatesBR().str_date_to_datetime(date_sup)
        # query parameters for request: filetype / series id / realtime start / realtime end
        query_params = {
            'series_id': data_source_id,
            'file_type': file_type,
            'api_key': api_key,
            'realtime_start': date_inf,
            'realtime_end': date_sup
        }
        # return fred data
        message_content = request('GET', url_root.format(
            data_source), params=query_params).content
        return HandlingObjects().literal_eval_data(message_content, "b'", "'")

    @property
    def non_farm_payroll(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_td = list()
        list_ser_1 = list()
        list_ser_2 = list()
        # payload
        dict_payload = {}
        # header
        dict_headers = {
        'Referer': 'https://www.google.com/',
        'Upgrade-Insecure-Requests': '1',
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36',
        'sec-ch-ua': '"Not/A)Brand";v="8", "Chromium";v="126", "Google Chrome";v="126"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'Cookie': 'ASP.NET_SessionId=dp3ziqlcjny45xqehrow1v2a; TEServer=TEIIS'
        }
        # requesting response
        resp_req = request(YAML_USA_MACRO['non_farm_payroll']['req_method'], 
                           YAML_USA_MACRO['non_farm_payroll']['url'], headers=dict_headers, 
                           data=dict_payload, 
                           verify=YAML_USA_MACRO['non_farm_payroll']['bl_verify'])
        # status code different from 2xx raise error
        resp_req.raise_for_status()
        # parsing html
        bs_html = BeautifulSoup(resp_req.content, 'html.parser')
        # * TABLE HISTORICAL DATA NON-FARM PAYROLL
        # headers
        list_th_0 = [
            td.get_text()
                .strip()
                .upper()
            for td in bs_html.find_all('th')
        ][:7]
        # data
        list_td1 = [
            td.get_text()
                .strip()
            for td in bs_html.find_all('td', class_='datatable-item datatable-item-positive')
        ]
        list_td2 = [
            td.get_text()
                .strip()
            for td in bs_html.find_all(
                'td', 
                style='text-align: left; padding-left: 10px; white-space: nowrap;'
            )
        ]
        list_td3 = [
            td.get_text()
                .strip()
            for td in bs_html.find_all('td', style='white-space: nowrap')
        ]
        list_td4 = [
            td.get_text()
                .strip()
            for td in bs_html.find_all('td', style='text-align: left')
        ]
        for i in range(len(list_td3)):
            list_td.extend([
                list_td2[i], list_td3[i], list_td4[i], list_td1[i+i*3], list_td1[i+1+i*3], 
                list_td1[i+2+i*3], list_td1[i+3+i*3]
            ])
        # creating list of dictionaries
        list_ser = HandlingDicts().pair_headers_with_data(
            list_th_0, 
            list_td
        )
        # serialized list into dataframe
        df_hist_nf_payroll = pd.DataFrame(list_ser)
        # * TABLE PAYROLLS AND USA LABOUR DATA
        for i in range(1, 3):
            #   table
            bs_table = bs_html.find_all('table')[i]
            #   headers
            if i == 1:
                list_th_1 = [
                    th.get_text()
                        .replace('\n', '')
                        .strip()
                        .upper() + str(i)
                    for th in bs_table.find_all('th')
                ]
            elif i == 2:
                list_th_2 = [
                    th.get_text()
                        .replace('\n', '')
                        .strip()
                        .upper() + str(i)
                    for th in bs_table.find_all('th')
                ]
            else:
                raise Exception('Table of payroll not supported for data treatment, ' 
                                + 'please revisit the parameter')
            #   data
            list_td = [
                td.get_text()
                    .replace('\n', '')
                    .strip()
                for td in bs_table.find_all('td')
            ]
            #   creating list of dictionaries
            if i == 1:
                list_ser_1 = HandlingDicts().pair_headers_with_data(
                    list_th_1, 
                    list_td
                )
            elif i == 2:
                list_ser_2 = HandlingDicts().pair_headers_with_data(
                    list_th_2, 
                    list_td
                )
                # print(list_ser_2)
            else:
                raise Exception('Table of payroll not supported for data treatment, ' 
                                + 'please revisit the parameter')
        # serialized list into dataframe
        df_payrolls = pd.DataFrame(list_ser_1)
        df_payrolls = df_payrolls[[c for c in list(df_payrolls.columns) 
                                   if (c not in list_th_0) and (c not in list_th_2)]]
        df_usa_labour_data = pd.DataFrame(list_ser_2)
        df_usa_labour_data = df_usa_labour_data[[c for c in list(
            df_usa_labour_data.columns) if (c not in list_th_1) and (c not in list_th_0)]]
        # looping through dataframes
        for i, df_ in enumerate([
            df_payrolls, 
            df_usa_labour_data
        ]):
            #   dropping rows that are empty
            df_.dropna(subset=list(df_.columns), how='all', inplace=True)
            #   removing duplicates
            df_.drop_duplicates(inplace=True)
            #   renaming columns
            df_.rename(columns={
                x: StrHandler().get_string_until_substr(
                    x, str(i+1)) for x in list(df_.columns)
            }, inplace=True)
            #   adding logging
            df_ = DBLogs().audit_log(
                df_, 
                YAML_USA_MACRO['non_farm_payroll']['url'],
                DatesBR().utc_from_dt(DatesBR().curr_date)
            )
        # adding logging to the last dataframe
        df_hist_nf_payroll = DBLogs().audit_log(
            df_hist_nf_payroll, 
            YAML_USA_MACRO['non_farm_payroll']['url'],
            DatesBR().utc_from_dt(DatesBR().curr_date)
        )
        # returning dataframes
        return df_hist_nf_payroll, df_payrolls, df_usa_labour_data

# print(USAMacro().fred_request('ceaca22bd10bd67a93fca0a5cd4aa578', 'series', 'GNPCA'))
# # output
# {'realtime_start': '2020-12-07', 'realtime_end': '2020-12-07', 'seriess': [{'id': 'GNPCA', 'realtime_start': '2020-12-07', 'realtime_end': '2020-12-07', 'title': 'Real Gross National Product', 'observation_start': '1929-01-01', 'observation_end': '2019-01-01', 'frequency': 'Annual',
#                                                                             'frequency_short': 'A', 'units': 'Billions of Chained 2012 Dollars', 'units_short': 'Bil. of Chn. 2012 $', 'seasonal_adjustment': 'Not Seasonally Adjusted', 'seasonal_adjustment_short': 'NSA', 'last_updated': '2020-07-30 07:57:33-05', 'popularity': 13, 'notes': 'BEA Account Code: A001RX\\n\\n'}]}

# print(USAMacro().fred_request(
#     'ceaca22bd10bd67a93fca0a5cd4aa578', 'series', 'DGS10'))



### DEALING WINDOWS OPERATING SYSTEM ###

import win32con
from ctypes import *
import win32gui


EnumWindows = windll.user32.EnumWindows
EnumWindowsProc = WINFUNCTYPE(c_bool, c_int, POINTER(c_int))
GetWindowText = windll.user32.GetWindowTextW
GetWindowTextLength = windll.user32.GetWindowTextLengthW
IsWindowVisible = windll.user32.IsWindowVisible
GetClassName = windll.user32.GetClassNameW
BringWindowToTop = windll.user32.BringWindowToTop
GetForegroundWindow = windll.user32.GetForegroundWindow


class DealingWindows:
    '''
    REFERENCE: makble.com/how-to-find-window-with-wildcard-in-python-and-win32gui
    DOCSTRING: MANIPULATE WINDOWS IN OS
    '''

    def __init__(self):
        self.titles = []

    def foreach_window(self, hwnd, lParam):
        '''
        DOCSTRING: SEARCH ALL WINDOWS OPENED
        INPUT: WINDOWS HANDLER
        OUTPUT: -
        '''
        if IsWindowVisible(hwnd):
            length = GetWindowTextLength(hwnd)
            classname = create_unicode_buffer(100 + 1)
            GetClassName(hwnd, classname, 100 + 1)
            buff = create_unicode_buffer(length + 1)
            GetWindowText(hwnd, buff, length + 1)
            self.titles.append((hwnd, buff.value.encode, classname.value,
                                windll.user32.IsIconic(hwnd)))

    def refresh_wins(self):
        '''
        DOCSTRING: REFRESH THE LIST OF WINDOWS TITLES OPENED
        INPUT: -
        OUTPUT: -
        '''
        self.titles = []
        EnumWindows(EnumWindowsProc(self.foreach_window), 0)

    def find_window(self, title_substring):
        '''
        DOCSTRING: SEARCH FOR WINDOW THAT CONTAINS THE TITLE SUBSTRING
        INPUT: TITLE SUBSTRING
        OUTPUT: WINDOW HANDLER
        '''
        self.refresh_wins()
        for item in self.titles:
            if title_substring.lower() in item[2].lower():
                return item[0]
        return False

    def close_window(self, window_handler):
        '''
        DOCSTRING: CLOSE THE WINDOW INDICATED BY WINDOW HANDLER
        INPUT: WINDOW HANDLER
        OUTPUT: -
        '''
        win32gui.PostMessage(window_handler, win32con.WM_CLOSE, 0, 0)



### DATABASE LOGS

from stpstone.settings._global_slots import YAML_GEN
from stpstone.cals.handling_dates import DatesBR


class DBLogs:

    def audit_log(self, df_, str_url, dt_last_update):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # adding audit columns
        df_[YAML_GEN['audit_log_cols']['url']] = str_url
        df_[YAML_GEN['audit_log_cols']['last_update']] = dt_last_update
        df_[YAML_GEN['audit_log_cols']['log_timestamp']] = DatesBR().utc_log_ts
        # returning dataframe
        return df_


# ROOT OF A FUNCTION WITH PYTHON

import numpy as np
from scipy.optimize import newton, fsolve


class Root:

    def bisection(self, f, a, b, epsilon):
        '''
        REFERENCES: PYTHON PROGRAMMING AND NUMERICAL METHODS, A GUIDE FOR ENGINEERS AND SCIENTS - 
            QINGKAI KONG, TIMMY SIAUW, ALEXANDRE M. BAYERN
        DOCSTRING: 
        INPUTS: FUNCTION (F), A, B (BOUNDARIES) AND EPSILON (TOLERANCE)
        OUTPUTS: FLOAT
        '''
        # approximates a root, R, of f bounded
        # by a and b to within tolerance
        # |f(m)| < epsilon with m being the midpoint
        # between a and b. Recursive implementation
        # check if a and b bound a root
        if np.sign(f(a)) == np.sign(f(b)):
            raise Exception('the scalars a and b do not bound a root')
        # get midpoint
        m = (a + b) / 2
        if np.abs(f(m)) < epsilon:
            #   stopping condition, report m as root
            return m
        elif np.sign(f(a)) == np.sign(f(m)):
            #   case where m is an improvement on a.
            #   make recursive call with a = m
            return self.bisection(f, m, b, epsilon)
        elif np.sign(f(b)) == np.sign(f(m)):
            #   case where m is an improvement on b.
            #   make recursive call with b = m
            return self.bisection(f, a, m, epsilon)

    def newton_raphson(self, f, x0, epsilon):
        '''
        DOCSTRING: NEWTHON RAPHSON METHOD TO OPTIMIZE ROOT-FINDING
        INPUTS: F(FUNCTION), X0 (FIRST ATTEMPT) AND EPSILON
        OUTPUTS: FLOAT
        '''
        return newton(f, x0, tol=epsilon)

    def fsolve(self, f, x0, epsilon):
        '''
        DOCSTRING: FSOLVE METHOD TO OPTIMIZE ROOT-FINDING
        INPUTS: F(FUNCTION), X0 (FIRST ATTEMPT) AND EPSILON
        OUTPUTS: FLOAT
        '''
        return fsolve(f, x0, xtol=epsilon)



### HANDLING API REQUESTS ###

# pypi.org libs
import pandas as pd
from requests.adapters import HTTPAdapter
from requests import Session, request
from requests.exceptions import ProxyError, ConnectTimeout, SSLError
from urllib3.util import Retry
from typing import Dict, Union, Any, List, Tuple
from random import shuffle
# local libs
from stpstone.settings._global_slots import YAML_SESSION
from stpstone.handling_data.dicts import HandlingDicts
from stpstone.loggs.create_logs import conditional_timeit


class ProxyServers:

    @property
    def proxy_scrape_free(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        resp_req = request(
            YAML_SESSION['proxy_scrape']['method'], 
            YAML_SESSION['proxy_scrape']['url'], 
        )
        resp_req.raise_for_status()
        json_proxies = resp_req.json()
        return [
            {
                'protocol': str(dict_['protocol']).lower(),
                'bl_alive': bool(dict_['alive']),
                'status': str(dict_['ip_data']['status']) if 'ip_data' in dict_ else '',
                'alive_since': float(dict_['alive_since']),
                'anonymity': str(dict_['anonymity']).lower(),
                'average_timeout': float(dict_['average_timeout']),
                'first_seen': float(dict_['first_seen']),
                'ip_data': str(dict_['ip_data']['as']) if 'ip_data' in dict_ else '',
                'ip_name': str(dict_['ip_data']['asname']) if 'ip_data' in dict_ else '',
                'timezone': str(dict_['ip_data']['timezone']) if 'ip_data' in dict_ else '',
                'continent': str(dict_['ip_data']['continent']) if 'ip_data' in dict_ else '',
                'continent_code': str(dict_['ip_data']['continentCode']) if 'ip_data' in dict_ else '',
                'country': str(dict_['ip_data']['country']) if 'ip_data' in dict_ else '',
                'country_code': str(dict_['ip_data']['countryCode']) if 'ip_data' in dict_ else '',
                'city': str(dict_['ip_data']['city']) if 'ip_data' in dict_ else '',
                'district': str(dict_['ip_data']['district']) if 'ip_data' in dict_ else '',
                'region_name': str(dict_['ip_data']['regionName']) if 'ip_data' in dict_ else '',
                'zip': str(dict_['ip_data']['zip']) if 'ip_data' in dict_ else '',
                'bl_hosting': bool(dict_['ip_data']['hosting']) if 'ip_data' in dict_ else '',
                'isp': str(dict_['ip_data']['isp']) if 'ip_data' in dict_ else '',
                'latitude': float(dict_['ip_data']['lat']) if 'ip_data' in dict_ else '',
                'longitude': float(dict_['ip_data']['lon']) if 'ip_data' in dict_ else '',
                'organization': str(dict_['ip_data']['org']) if 'ip_data' in dict_ else '',
                'proxy': str(dict_['proxy']),
                'ip': str(dict_['ip']),
                'port': int(dict_['port']),
                'bl_ssl': bool(dict_['ssl']),
                'timeout': float(dict_['timeout']),
                'times_alive': float(dict_['times_alive']),
                'times_dead': float(dict_['times_dead']),
                'ratio_times_alive_dead': float(dict_['times_alive'] / dict_['times_dead']) 
                    if 'times_alive' in dict_ and 'times_dead' in dict_ and dict_['times_dead'] != 0 
                    else 0,
                'uptime': float(dict_['uptime'])
            } for dict_ in json_proxies['proxies']
        ]

    @property
    def available_proxies(self):
        return self.proxy_scrape_free


class ReqSession(ProxyServers):

    def __init__(self, bl_proxy:bool=True, dict_proxies:Union[Dict[str, str], None]=None, 
                 int_retries:int=10, int_backoff_factor:int=1, bl_alive:bool=True,
                 list_anonimity_value:Union[List[str], str, None]=['anonymous', 'elite'], 
                 str_protocol:str='http', str_continent_code:Union[str, None]=None, 
                 str_country_code:Union[str, None]=None, bl_ssl:Union[bool, None]=None, 
                 float_ratio_times_alive_dead:Union[float, None]=0.02,
                 float_min_timeout:Union[float, None]=600, bl_use_timer:bool=False,
                 list_status_forcelist:list=[429, 500, 502, 503, 504]) -> None:
        '''
        DOCSTRING: SESSION CONFIGURATION
        INPUTS: 
            - URL:STR
            - PROXIES:DICT (NONE AS DEFAULT) 
                . FORMAT: {'http': 'http://127.0.0.1:8080', 'https': 'http://127.0.0.1:8080'}
            - RETRIES:INT (10 AS DEFAULT)
            - BACKOFF_FACTOR:INT (1 AS DEFAULT)
            - STATUS_FORCELIST:LIST (429, 500, 502, 503, 504 AS DEFAULT)
        OUTPUTS: SESSION
        '''
        self.bl_proxy = bl_proxy
        self.int_retries = int_retries
        self.int_backoff_factor = int_backoff_factor
        self.bl_alive = bl_alive
        self.list_anonimity_value = list_anonimity_value
        self.str_protocol = str_protocol
        self.str_continent_code = str_continent_code
        self.str_country_code = str_country_code
        self.bl_ssl = bl_ssl
        self.float_ratio_times_alive_dead = float_ratio_times_alive_dead
        self.float_min_timeout = float_min_timeout
        self.bl_use_timer = bl_use_timer
        self.list_status_forcelist = list_status_forcelist
        self.proxy = self.get_proxy if bl_proxy == True else None
        self.dict_proxy = dict_proxies if dict_proxies is not None else (
            self._dict_proxy(self.proxy['ip'], self.proxy['port']) 
            if self.proxy is not None else None
        )
        self.session = self.configure_session(self.dict_proxy, self.int_retries, 
                                              self.int_backoff_factor)
        self.ip_infos = self.ip_infos(self.session, bl_return_availability=False)

    def _dict_proxy(self, str_ip:str, int_port:int) -> Dict[str, str]:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return {
            'http': 'http://{}:{}'.format(str_ip, str(int_port)), 
            'https': 'http://{}:{}'.format(str_ip, str(int_port))
        }

    def configure_session(self, dict_proxy:Union[Dict[str, str], None]=None, 
                          int_retries:int=10, int_backoff_factor:int=1) -> Session:
        '''
        DOCSTRING: CONFIGURES AN HTTP SESSION WITH RETRY MECHANISM AND EXPONENTIAL BACKOFF
        INPUTS: NONE
        OUTPUTS: CONFIGURED HTTP SESSION OBJECT
        OBS:
            1. RETRY_STRATEGY OVERVIEW:
                - TOTAL: TOTAL NUMBER OF RETRIES
                - BACKOFF_FACTOR: EXPONENTIAL BACKOFF FACTOR
                    . CALCULATED AS THE DEALAY BEFORE THE NEXT RETRY
                    . DELAY = BACKOFF_FACTOR * (2 ** (RETRY_NUMBER - 1))
                    . AFTER THE 1ST RETRY: DELAY = 1 * 2**0 = 1 SECOND
                    . AFTER THE 2ND RETRY: DELAY = 1 * 2**1 = 4 SECONDS
                    . IN THE AFORE EXAMPLE THE BACKOFF FACTOR IS 1
                - STATUS_FORCELIST: LIST OF STATUS CODES TO RETRY
            2. SESSION OBJECT OVERVIEW:
                - MOUNT: MOUNTS THE RETRY STRATEGY TO THE SESSION, WITH THE GIVEN ADAPTER
                - SESSION OBJECTS HAVE METHODS AS .GET() AND .POST()
        '''
        retry_strategy = Retry(
            total=int_retries,
            backoff_factor=int_backoff_factor,
            status_forcelist=self.list_status_forcelist
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session = Session()
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        if dict_proxy is not None:
            session.proxies.update(dict_proxy)
        return session

    def ip_infos(self, session:Session, bl_return_availability:bool=False, 
                 tup_timeout:Tuple[int, int]=(5,5)) -> Union[List[Dict[str, Any]], None]:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        dict_payload = {}
        dict_headers = {
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
            'accept-language': 'en-US,en;q=0.9,pt;q=0.8,es;q=0.7',
            'cache-control': 'max-age=0',
            'priority': 'u=0, i',
            'sec-ch-ua': '"Not A(Brand";v="8", "Chromium";v="132", "Google Chrome";v="132"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Windows"',
            'sec-fetch-dest': 'document',
            'sec-fetch-mode': 'navigate',
            'sec-fetch-site': 'none',
            'sec-fetch-user': '?1',
            'upgrade-insecure-requests': '1',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36'
        }
        resp_req = session.get(YAML_SESSION['ipinfos']['url'], headers=dict_headers, 
                                    data=dict_payload, timeout=tup_timeout)
        resp_req.raise_for_status()
        if bl_return_availability == True:
            return True
        else:
            return resp_req.json()
    
    def test_proxy(self, str_ip:str, int_port:int) -> bool:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        try:
            session = self.configure_session(
                dict_proxy={
                    'http': 'http://{}:{}'.format(str_ip, str(int_port)), 
                    'https': 'http://{}:{}'.format(str_ip, str(int_port))
                },
                int_retries=0,
                int_backoff_factor=0
            )
            return self.ip_infos(session, bl_return_availability=True)
        except (ProxyError, ConnectTimeout, SSLError):
            return False

    @property
    def _proxies(self) -> List[Dict[str, Union[str, int]]]:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        list_ser = self.available_proxies
        # filtering proxies
        for k_filt, v_filt, str_strategy in [
            ('bl_alive', self.bl_alive, 'equal'),
            ('anonymity', self.list_anonimity_value, 'isin'),
            ('protocol', self.str_protocol, 'equal'),
            ('bl_ssl', self.bl_ssl, 'equal'),
            ('ratio_times_alive_dead', self.float_ratio_times_alive_dead, 'greater_than_or_equal_to'),
            ('timeout', self.float_min_timeout, 'greater_than_or_equal_to'),
            ('continent_code', self.str_continent_code, 'equal'),
            ('country_code', self.str_country_code, 'equal')
        ]:
            if v_filt is not None:
                try:
                    list_ser = HandlingDicts().filter_list_ser(
                        list_ser, 
                        k_filt, 
                        v_filt,
                        str_filter_type=str_strategy
                    )
                except KeyError as e:
                    print(list_ser)
                    raise Exception(e)
        return list_ser

    @property
    def get_proxy(self) -> Union[Dict[str, Any], None]:
        '''
        DOCSTRING: RETRIEVES A VALID PROXY FROM THE FILTERED LIST, APPLYING THE TEST PROXY METHOD
        INPUTS: -
        OUTPUTS: DICT
        '''
        @conditional_timeit(bl_use_timer=self.bl_use_timer)
        def retrieve_proxy():
            list_ser = self._proxies
            shuffle(list_ser)
            for dict_proxy in list_ser:
                str_ip = dict_proxy['ip']
                int_port = dict_proxy['port']
                if all([x is not None for x in [str_ip, int_port]]) == True:
                    if self.test_proxy(str_ip, int_port):
                        return {'ip': str_ip, 'port': int_port}
            return None
        return retrieve_proxy()



### API TO GRANT ACCESS TO B3 MARGIN SIMULATOR ###

from requests import request
from stpstone.handling_data.json import JsonFiles


class MarginSimulatorB3:

    def __init__(self, data_payload_portfolio, token='79a4413f55d7d982b61c669e6dd35eea',
                 hostname_api_margin_simulator_b3='https://simulador.b3.com.br/api/cors-app'):
        '''
        INPUTS: DATA PAYLOAD PORTFOLIO EXEMPLE - data_payload_portfolio = [
                {
                    'Security': {'symbol': 'ABCBF160'}
                    'SecurityGroup': {'positionTypeCode': 0},
                    'Position': {'longQuantity':100,'shortQuantity': 0,'longPrice': 0,'shortPrice': 0}
                },
                {
                    'Security': {'symbol': 'ABCBF179'},
                    'SecurityGroup': {'positionTypeCode': 0},
                    'Position': {'longQuantity': 100,'shortQuantity': 0,'longPrice': 0,'shortPrice': 0}
                },
                {
                    'Security': {'symbol': 'ABCBF182'},
                    'SecurityGroup': {'positionTypeCode': 0},
                    'Position': {'longQuantity': 0,'shortQuantity':200,'longPrice': 0,'shortPrice': 0}
                }
            ]
        '''
        self.data_payload_portfolio = data_payload_portfolio
        self.token = token,
        self.hostname_api_margin_simulator_b3 = hostname_api_margin_simulator_b3

    @property
    def total_deficit_surplus(self, method='POST', app='/web/V1.0/RiskCalculation',
                              value_liquidity_resource=4700000000, bl_verify=False,
                              bl_parse_dict_payload_data=True):
        '''
        REFERENCES: https://simulador.b3.com.br/
        DOCSTRING: TOTAL DEFICIT SURPLUS B3 MARGIN CALL CALCULATION
        INPUTS: METHOD (DEFAULT), KEY HEADER (DEFAULT), URL AUTHENTIFICATION
            HEADER (DEFAULT)
        OUTPUTS: STRING
        '''
        # requesting authorization authheader
        dict_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        # payload
        dict_payload = {
            'ReferenceData': {'referenceDataToken': '{}'.format(self.token)},
            'LiquidityResource': {'value': value_liquidity_resource},
            'RiskPositionList': self.data_payload_portfolio
        }
        # check wheter it is needed to parse the params dictionary
        if bl_parse_dict_payload_data == True:
            if dict_payload != None:
                dict_payload = JsonFiles().dict_to_json(dict_payload)
        # resquet host REST information
        resp_req = request(method=method, url=self.hostname_api_margin_simulator_b3 + app,
                           headers=dict_headers, data=dict_payload, verify=bl_verify)
        # raises exception when not a 2xx response
        resp_req.raise_for_status()
        # getting authheader
        return resp_req.json()



# OPTION PRCING FORMULAS, FOR EUROPEAN TYPE

import numpy as np
from math import pi
from scipy.optimize import fsolve, minimize
from functools import lru_cache
from stpstone.quantitative_methods.prob_distributions import NormalDistribution
from stpstone.quantitative_methods.regression import NonLinearEquations


class InitialSettings:

    def set_parameters(self, *params, opt_type='call'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # check wheter is a call or put option, in case the type is neither of the former raise
        #   error
        if opt_type not in ['call', 'put']:
            raise Exception('Option ought be a call or a put')
        # initial parameters
        list_params = [param[0] if isinstance(param, list) == True else param for
                       param in params]
        # return parameters
        return [float(param) for param in list_params if isinstance(param, str) == False]


class BlackScholesMerton(InitialSettings):
    '''
    REFERENCES: https://brilliant.org/wiki/black-scholes-merton/
    '''

    def d1(self, s, k, b, t, sigma, q):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: D1 OF UNDERLYING OPTION
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY), B (COST OF CARRY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET) AND Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, b, t, sigma, q = self.set_parameters(s, k, b, t, sigma, q)
        # return d1 probability
        return (np.log(s / k) + (b + sigma ** 2 / 2) * t) / (sigma * np.sqrt(t))

    def d2(self, s, k, b, t, sigma, q):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: D2 OF UNDERLYING OPTION (MONEYNESS)
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY), B (COST OF CARRY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET) AND Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, b, t, sigma, q = self.set_parameters(s, k, b, t, sigma, q)
        # return d2 probability
        return self.d1(s, k, b, t, sigma, q) - sigma * np.sqrt(t)

    def general_opt_price(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: CALL/PUT PRICE OF AN UNDERLYING ASSET
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), 
            T (TIME TO MATURITY), SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), 
            B (COST OF CARRY - R FOR STOCK OPTION, R - Q FOR STOCK 
            OPTION WITH CONTINUOUS DIVIDEND YIELD, 0 FOR FUTURES, 0 AND R 0 FOR MARGINED FUTURES 
            OPTIONS, AND R - RF FOR CURRENCY OPTION MODEL) AND OPTION STYLE (CALL/PUT)
        OUTPUTS: CALL PRICE
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return option price
        if opt_type == 'call':
            return s * np.exp((b - r) * t) * \
                NormalDistribution().cdf(self.d1(s, k, b, t, sigma, q)) \
                - k * np.exp(-r * t) * NormalDistribution().cdf(BlackScholesMerton(
                ).d2(s, k, b, t, sigma, q))
        elif opt_type == 'put':
            return k * np.exp(-r * t) * NormalDistribution().cdf(
                -self.d2(s, k, b, t, sigma, q)) - s * np.exp((b - r) * t) * NormalDistribution().cdf(
                -self.d1(s, k, b, t, sigma, q))


class Greeks(BlackScholesMerton):
    '''
    REFERENCES: https://www.macroption.com/option-greeks-excel/, https://en.wikipedia.org/wiki/Greeks_(finance)
    INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY), R (INTEREST RATE),
        SIGMA (VOLATILITY OF UNDERLYING ASSET) AND Q (DIVIDEND YIELD)
    '''

    def delta(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF THE THEORETICAL OPTION VALUE WITH RESPECT TO CHANGES IN THE
            UNDERLYING ASSET'S PRICE - FIRST DERIVATIVE OF THE OPTION VALUE WITH RESPECT TO THE
            UNDERLYING INSTRUMENT'S PRICE S
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD) B (COST OF CARRY) 
            AND OPTION STYLE (CALL/PUT)
        OUTPUTS:
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return np.exp((b - r) * t) * NormalDistribution().cdf(self.d1(
                s, k, b, t, sigma, q))
        elif opt_type == 'put':
            return np.exp((b - r) * t) * (NormalDistribution().cdf(
                self.d1(s, k, b, t, sigma, q)) - 1)

    def future_delta_from_spot_delta(self, delta, b, t):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: IN SOME MARKETS IT IS OPTIONAL TO HEDGE WITH THE STOCK ITSELF OR, ALTERNATIVELY, 
            HEDGE WITH THE STOCK FUTURES - IN THE CASE WHERE ONE HEDGE WITH A FORWARD CONTRACT WITH 
            THE SAME EXPIRATION AS THE OPTION THE FORMULA ALSO HOLDS TRUE - THIS IS PARTICULARLY 
            USEFUL IN THE FX MARKET, WHERE TYPICALLY CAN BE CHOOSEN BETWEEN HEDGING WITH THE 
            CURRENCY SPOT OR ALTERNATIVELY A FORWARD WITH EXPIRATION MATCHING THE OPTION EXPIRATION
        INPUTS:
        OUTPUTS:
        '''
        return delta * np.exp(-b * t)

    def strike_from_delta(self, s, r, t, sigma, q, b, delta, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF THE THEORETICAL OPTION VALUE WITH RESPECT TO CHANGES IN THE
            UNDERLYING ASSET'S PRICE - FIRST DERIVATIVE OF THE OPTION VALUE WITH RESPECT TO THE
            UNDERLYING INSTRUMENT'S PRICE S
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY) 
            AND OPTION STYLE (CALL/PUT)
        OUTPUTS:
        '''
        # initial parameters
        s, r, t, sigma, q, b = self.set_parameters(
            s, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return s * np.exp(-NormalDistribution().inv_cdf(delta * np.exp((r - b) * t))
                              * sigma * t ** 0.5 + (b + sigma ** 2 / 2.0) * t)
        elif opt_type == 'put':
            return s * np.exp(NormalDistribution().inv_cdf(delta * np.exp((r - b) * t))
                              * sigma * t ** 0.5 + (b + sigma ** 2 / 2.0) * t)

    def gamma(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: DELTA SENSITIVITY TO SMALL CHANGES IN THE UNDERLYING PRICE
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY) 
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(s, k, r, t, sigma, q, b)
        # return greek
        return NormalDistribution().pdf(self.d1(s, k, b, t, sigma, q)) * np.exp((b - r) * t) \
            / (s * sigma * t ** 0.5)

    def saddle_gamma(self, k, r, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: CRITICAL POINT OF GAMMA - LOWER BOUNDARY
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY) 
        OUTPUTS: FLOAT
        '''
        # initial parameters
        k, r, sigma, q, b = self.set_parameters(k, r, sigma, q, b)
        # return greek
        return np.sqrt((np.exp(1) / pi) * ((2 * b - r) / sigma ** 2 + 1)) / k

    def gamma_p(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: PERCENTAGE CHAGENS IN DELTA FOR PERCENTAGE CHANGES IN THE UNDERLYING (GAMMA 
            PERCENT)
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY) 
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(s, k, r, t, sigma, q, b)
        # return greek
        return NormalDistribution().pdf(self.d1(s, k, b, t, sigma, q)) * np.exp((b - r) * t) \
            / (100.0 * sigma * t ** 0.5)

    def theta(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        DOCSTRING: SENSITIVITY  MEASUREMENT OF TIME  BEFORE EXPIRATION DATE (IT WILL LOSE VALUE
            PRICED INTO THE EXTRINSIC VALUE OVER TIME) - GAUGE HOW MUCH VALUE AN OPTION LOSES ON
            A DAILY BASIS
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), OPTION CALL OR PUT AND T
            (NUMBER OF DAY PER YEAR, GENERALLY 365 OR 252)
        OUTPUTS:
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return -(s * np.exp((b - r) * t) * NormalDistribution().pdf(self.d1(
                s, k, b, t, sigma, q)) * sigma) / (2.0 * t ** 0.5) - (b - r) * s * np.exp(
                (b - r) * t) * NormalDistribution().cdf(self.d1(s, k, b, t, sigma, q)) \
                - r * k * \
                np.exp(-r * t) * \
                NormalDistribution().cdf(self.d2(s, k, b, t, sigma, q))
        elif opt_type == 'put':
            return -(s * np.exp((b - r) * t) * NormalDistribution().pdf(self.d1(
                s, k, b, t, sigma, q)) * sigma) / (2.0 * t ** 0.5) + (b - r) * s * np.exp(
                (b - r) * t) * NormalDistribution().cdf(-self.d1(s, k, b, t, sigma, q)) \
                + r * k * \
                np.exp(-r * t) * \
                NormalDistribution().cdf(-self.d2(s, k, b, t, sigma, q))

    def vega(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: SENSITIVITY MEASUREMENT OF VOLATILITY OVER TIME
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: VEGA
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(s, k, r, t, sigma, q, b)
        # return greek
        return s * np.exp((b - r) * t) * NormalDistribution().pdf(self.d1(
            s, k, b, t, sigma, q)) * np.sqrt(t)

    def vega_local_maximum(self, k, t, sigma, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: LOCAL MAXIMUM OF VEGA
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: VEGA
        '''
        # initial parameters
        k, t, sigma, b = self.set_parameters(k, t, sigma, b)
        # return greek
        return k * np.exp(-b + sigma ** 2 / 2.0) * t

    def strike_maximizes_vega(self, s, t, sigma, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: STRIKE THAT MAXIMIZES VEGA, GIVEN THE ASSET PRICE
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: VEGA
        '''
        # initial parameters
        s, t, sigma, b = self.set_parameters(s, t, sigma, b)
        # return greek
        return s * np.exp(b + sigma ** 2 / 2.0) * t

    def time_to_maturity_maximum_vega(self, s, k, r, sigma, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: TIME TO MATURITY WHEN VEGA IS THE GREATEST
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: VEGA
        '''
        # initial parameters
        s, k, r, sigma, b = self.set_parameters(s, k, r, sigma, b)
        # return greek
        return 2 * (1.0 + (1.0 + (8.0 * r / sigma ** 2 + 1.0) * np.log(s / k) ** 2) ** 0.5) \
            / (8.0 * r + sigma ** 2)

    def vega_global_maximum(self, k, r, sigma, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: TIME TO MATURITY WHEN VEGA IS THE GREATEST
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: VEGA
        '''
        # initial parameters
        k, r, sigma, b = self.set_parameters(k, r, sigma, b)
        # global maximum time for vega
        t = 1 / (2.0 * r)
        # stock price at the maximum time
        s = k * np.exp((-b + sigma ** 2 / 2.0) * t)
        # vega at t
        vega = k / (2.0 * (r * np.exp(1) * pi) ** 0.5)
        # return greek
        return {
            't_max_global_vega': t,
            's_max_global_vega': s,
            'vega_max_global': vega
        }

    def vega_gamma_relationship(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RELATIONSHIP BETWEEN VEGA AND GAMMA - RETURNS VEGA
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(s, k, r, t, sigma, q, b)
        # return greek
        return self.gamma(s, k, r, t, sigma, q, b) * sigma * s ** 2 * t

    def vega_delta_relationship(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RELATIONSHIP BETWEEN VEGA AND GAMMA - RETURNS VEGA
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b, opt_type = self.set_parameters(s, k, r, t, sigma, q, b,
                                                                 opt_type)
        # return greek
        return np.exp((b - r) * t) * NormalDistribution().pdf(NormalDistribution().inv_cdf(
            np.exp((r - b) * t) * np.abs(self.delta(s, k, r, t, sigma, q, b, opt_type)))) \
            / (s * sigma * t ** 0.5)

    def vega_p(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: VEGA PERCENTUAL CHANGE
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B(COST OF CARRY) AND
            OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(s, k, r, t, sigma, q, b)
        # return greek
        return sigma / 10.0 * s * np.exp((b - r) * t) * NormalDistribution().pdf(
            self.d1(s, k, b, t, sigma, q)) * t ** 0.5

    def vega_elasticity(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: VEGA PERCENTUAL CHANGE
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B(COST OF CARRY) AND
            OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b, opt_type = self.set_parameters(s, k, r, t, sigma, q, b,
                                                                 opt_type)
        # return greek
        return self.vega(s, k, r, t, sigma, q, b) * sigma \
            / self.general_opt_price(s, k, r, t, sigma, q, b, opt_type)

    def rho(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        DOCSTRING: OPTION'S SENSITIVITY TO SMALL CHANGES IN THE RISK-FREE INTEREST RATE
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), OPTION CALL OR PUT AND T
            (NUMBER OF DAY PER YEAR, GENERALLY 365 OR 252)
        OUTPUTS:
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return t * k * np.exp(- r * t) * NormalDistribution().cdf(self.d2(s, k, b, t, sigma, q))
        elif opt_type == 'put':
            return -t * k * np.exp(- r * t) * NormalDistribution().cdf(-self.d2(s, k, b, t, sigma, q))

    def lambda_greek(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        DOCSTRING: 
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD) OPTION CALL OR PUT
        OUTPUTS: LAMBDA
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        return self.delta(s, k, r, t, sigma, q, opt_type) * s / \
            self.general_opt_price(s, k, r, t, sigma, q, b, opt_type)

    def vanna(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF VEGA ACCORDING TO SPOT, OR DELTA TO VOLATILITY (SIGMA)
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD) B (COST OF CARRY) 
            AND OPTION STYLE (CALL/PUT)
        OUTPUTS:
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        return -np.exp((b - r) * t) * self.d2(s, k, b, t, sigma, q) * NormalDistribution().pdf(
            self.d1(s, k, b, t, sigma, q)) / sigma

    def vanna_vol(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF VANNA ACCORDING TO VOLATILITY (SIGMA) (SECOND-ORDER PARTIAL 
            DERIVATIVE OF DELTA WITH RESPECT TO VOLATILITY)
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD) B (COST OF CARRY) 
            AND OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        return self.vanna(s, k, r, t, sigma, q, b, opt_type) * (1.0 / sigma) \
            * (self.d1(s, k, b, t, sigma, q) * self.d2(s, k, b, t, sigma, q)
                - self.d1(s, k, b, t, sigma, q) / self.d2(s, k, b, t, sigma, q) - 1.0)

    def charm(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF DELTA TO TIME, ALSO KNOWN AS CHARM OR DELTA BLEED
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD) B (COST OF CARRY) 
            AND OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return -np.exp((b - r) * t) * (NormalDistribution().pdf(
                self.d1(s, k, b, t, sigma, q)) * (b / (sigma * t ** 0.5) - self.d2(
                    s, k, b, t, sigma, q) / (2.0 * t)) + (b - r) * NormalDistribution().cdf(
                        self.d1(s, k, b, t, sigma, q)))
        elif opt_type == 'put':
            return -np.exp((b - r) * t) * (NormalDistribution().pdf(
                self.d1(s, k, b, t, sigma, q)) * (b / (sigma * t ** 0.5) - self.d2(
                    s, k, b, t, sigma, q) / (2.0 * t)) - (b - r) * NormalDistribution().cdf(
                        -self.d1(s, k, b, t, sigma, q)))

    def zomma(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF GAMMA TO IMPLIED VOLATILITY, ALSO KNOWN AS ZOMMA
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            GAMMA (OR GAMMA_P)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return self.gamma(s, k, r, t, sigma, q, b) * (self.d1(s, k, b, t, sigma, q) * self.d2(
            s, k, b, t, sigma, q) - 1) / sigma

    def zomma_p(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: SPEED PERCENTAGE
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            GAMMA (OR GAMMA_P)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return self.zomma(s, k, r, t, sigma, q, b) * s / 100

    def speed(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF GAMMA TO SPOT PRICE, ALSO KNOWN AS SPEED
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            GAMMA (OR GAMMA_P)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return -self.gamma(s, k, r, t, sigma, q, b) * (1.0 + self.d1(s, k, b, t, sigma, q)
                                                       / (sigma * t ** 0.5)) / s

    def speed_p(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: SPEED PERCENTAGE
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            GAMMA (OR GAMMA_P)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return self.speed(s, k, r, t, sigma, q, b) * s / 100

    def color(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF GAMMA TO TIME, ALSO KNOWN AS COLOR
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            GAMMA (OR GAMMA_P)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return -self.gamma(s, k, r, t, sigma, q, b) * (r - b + b * self.d1(
            s, k, b, t, sigma, q) / (sigma * t ** 0.5) + (1.0 - self.d1(
                s, k, b, t, sigma, q) * self.d2(s, k, b, t, sigma, q)) / (2.0 * t))

    def color_p(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF GAMMA TO TIME, ALSO KNOWN AS COLOR
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            GAMMA (OR GAMMA_P)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return -self.gamma_p(s, k, r, t, sigma, q, b) \
            * (r - b + b * self.d1(s, k, b, t, sigma, q)
               / (sigma * t ** 0.5) + (1.0 - self.d1(s, k, b, t, sigma, q)
                                       * self.d2(s, k, b, t, sigma, q)) / (2.0 * t))

    def vomma(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: VEGA PERCENTUAL CONVEXITY, OR THE SENSITIVITY OF CHANGES IN IMPLIED VOLATILITY
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            GAMMA (OR GAMMA_P)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return self.vega(s, k, r, t, sigma, q, b) * self.d1(s, k, b, t, sigma, q) \
            * self.d2(s, k, b, t, sigma, q) / sigma

    def vomma_p(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: VEGA CONVEXITY, OR THE SENSITIVITY OF CHANGES IN IMPLIED VOLATILITY
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            GAMMA (OR GAMMA_P)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return self.vega_p(s, k, r, t, sigma, q, b) * self.d1(s, k, b, t, sigma, q) \
            * self.d2(s, k, b, t, sigma, q) / sigma

    def vomma_positive_outside_interval(self, s_k, r, t, sigma, q, b, bl_spot=True):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: INTERVAL BEYOND WHICH VOMMA BEGINS TO REGISTER POSITIVE VALUES -->
            INTERVAL TO SPOT OR STRIKE VALUES
        INPUTS: S (SPOT PRICE) OR K (STRIKE) (S_K), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            GAMMA (OR GAMMA_P)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # sign of cost of carry, according to s_k being a spot or strike value
        if bl_spot == True:
            sign_ = 1
        else:
            sign_ = -1
        # return vomma 0 boundaries
        return {
            'lower_boundary': s_k * np.exp((sign_ * b - sigma ** 2 / 2.0) * t),
            'upper_boundary': s_k * np.exp((sign_ * b + sigma ** 2 / 2.0) * t)
        }

    def ultima(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: VOMMA'S SENSITIVITY TO A CHANGE IN VOLATITLITY
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return self.vomma(s, k, r, t, sigma, q, b) / sigma \
            * (self.d1(s, k, b, t, sigma, q) * self.d2(s, k, b, t, sigma, q)
               - self.d1(s, k, b, t, sigma, q) / self.d2(s, k, b, t, sigma, q)
                - self.d2(s, k, b, t, sigma, q) / self.d1(s, k, b, t, sigma, q) - 1.0)

    def d_vega_d_time(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: VEGA'S SENSITIVITY TO A CHANGE IN TIME
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return self.vega(s, k, r, t, sigma, q, b) * (r - b + b * self.d1(
            s, k, b, t, sigma, q) / (sigma * t ** 0.5) - (1.0 + self.d1(
                s, k, b, t, sigma, q) * self.d2(s, k, b, t, sigma, q)) / (2.0 * t))

    def variance_vega(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: BSM'S FORMULA SENSITIVITY TO A SMALL CHANGE IN THE VARIANCE OF THE UNDERLYING 
            ASSET'S INSTANTENEOUS RATE OF RETURN
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return self.vega(s, k, r, t, sigma, q, b) / (2 * sigma)

    def variance_vanna(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: CHANGE IN DELTA FOR A CHANGE IN THE VARIANCE
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return - s * np.exp((b - r) * t) * self.pdf(self.d1(s, k, b, t, sigma, q)) * self.d2(
            s, k, b, t, sigma, q) / (2.0 * sigma)

    def variance_vomma(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: VARIANCE VEGA'S SENSITIVITY TO A SMALL CHANGE IN THE VARIANCE
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return s * np.exp((b - r) * t) * t ** 0.5 / (4.0 * sigma ** 3) * self.pdf(self.d1(
            s, k, b, t, sigma, q)) * (self.d1(s, k, b, t, sigma, q) * self.d2(
                s, k, b, t, sigma, q) - 1.0)

    def variance_ultima(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: VARIANCE ULTIMA IS THE THIRD DERIVATIVE OF BSM'S MODEL WITH RESPECT TO VARIANCE
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return s * np.exp((b - r) * t) * t ** 0.5 / (8 * sigma ** 5) * self.pdf(self.d1(
            s, k, b, t, sigma, q)) * ((self.d1(s, k, b, t, sigma, q) * self.d2(
                s, k, b, t, sigma, q) - 1.0) * (self.d1(s, k, b, t, sigma, q) * self.d2(
                    s, k, b, t, sigma, q) - 3.0) - (self.d1(s, k, b, t, sigma, q) ** 2 + self.d2(
                        s, k, b, t, sigma, q) ** 2))

    def dbsm_dohm(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RATE OF CHANGE OF BSM'S MODEL REGARDING OHM (STANDARD DEVIATION - TIME)
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return s * NormalDistribution().pdf((np.log(s / k) + t * sigma ** 2 / 2.0)
                                            / (sigma * t ** 0.5))

    def driftless_theta(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: TIME DECAY WITHOUT TAKING INTO ACCOUNT THE DRIFT OF THE UNDERLYING OR DISCOUNTING 
            - THE DRIFTLESS THETA THEREBY ISOLATES THE EFFECT TIME DECAY HAS ON UNCERTAINTY, 
            ASSUMING CONSTANT VOLATILITY - UNCERTAINTY AFFECTS THE OPTION THROUGH BOTH TIME AND 
            VOLATILITY, SINCE THE LATTER IS A MEASURE OF UNCERTAINTY DURING AN INFINITESIMAL TIME 
            PERIOD
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b)
        # return greek
        return -s * NormalDistribution().pdf(self.d1(s, k, b, t, sigma, q)) * sigma / (
            2.0 * t ** 0.5)

    def theta_vega_relationship(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RELATIONSHIP BETWEEN THETA AND VEGA - RETURNS THETA
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(s, k, r, t, sigma, q, b)
        # return greek
        return -self.gamma(s, k, r, t, sigma, q, b) * sigma / (2.0 * t)

    def bleed_offset_volatility(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: IT MEASURES HOW MUCH THE VOLATILITY MUST INCREASE TO OFFSET THE THETA-BLEED/TIME
            DECAY - IN THE CASE OF POSITIVE THETA, ONE CAN ACTUALLY HAVE NEGATIVA OFFSET VOLATILITY 
            - DITM EUROPEAN OPTIONS CAN HAVE POSITIVE THETA, AND IN THIS CASE THE OFFSET VOLATILITY 
            WILL BE NEGATIVE
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(s, k, r, t, sigma, q, b)
        # return greek
        return self.theta(s, k, r, t, sigma, q, b) / self.vega(s, k, r, t, sigma, q, b)

    def theta_gamma_relationship_driftless(self, s, k, r, t, sigma, q, b):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RELATIONSHIP BETWEEN DRIFTLESS GAMMA AND THETA - RETURNS DRIFTLESS THETA
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(s, k, r, t, sigma, q, b)
        # return greek
        return -2.0 * self.driftless_theta(s, k, r, t, sigma, q, b) / (
            s ** 2 * sigma ** 2)

    def phi(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: OPTION SENSITIVITY TO A CHANGE IN THE DIVIDEND YIELD (PHI, ALSO KNOWN AS RHO-2), 
            OR THE FOREIGN INTEREST RATE IN THE CASE OF A CURRENCY OPTION
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return -t * s * np.exp((b - r) * t) * NormalDistribution().cdf(self.d1(
                s, k, b, t, sigma, q))
        elif opt_type == 'put':
            return t * s * np.exp((b - r) * t) * NormalDistribution().cdf(-self.d1(
                s, k, b, t, sigma, q))

    def carry_rho(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: OPTION'S SENSITIVITY TO A SMALL CHANGE IN THE COST-OF-CARRY RATE
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return t * s * np.exp((b - r) * t) * NormalDistribution().cdf(self.d1(
                s, k, b, t, sigma, q))
        elif opt_type == 'put':
            return -t * s * np.exp((b - r) * t) * NormalDistribution().cdf(-self.d1(
                s, k, b, t, sigma, q))

    def risk_neutral_prob_itm(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: RISK-NEUTRAL PROBABILITY FOR ENDING UP ITM AT MATURITY
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return NormalDistribution().cdf(self.d2(s, k, b, t, sigma, q))
        elif opt_type == 'put':
            return NormalDistribution().cdf(-self.d2(s, k, b, t, sigma, q))

    def strike_given_risk_neutral_prob(self, s, k, r, t, sigma, q, b, p, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING:
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD), B (COST OF CARRY), 
            P (RISK-NEUTRAL PROBABILITY) AND opt_type (OPTION STYLE)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b, p = self.set_parameters(
            s, k, r, t, sigma, q, b, p, opt_type)
        # return greek
        if opt_type == 'call':
            return s * np.exp(-NormalDistribution().inv_cdf(p) * sigma * t ** 0.5
                              + (b - sigma ** 2 / 2.0) * t)
        elif opt_type == 'call':
            return s * np.exp(NormalDistribution().inv_cdf(p) * sigma * t ** 0.5
                              + (b - sigma ** 2 / 2.0) * t)

    def d_zeta_d_vol(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: ZETA'S SENSITIVITY TO A SMALL CHANGE IN THE IMPLIED VOLATILITY
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return -NormalDistribution().pdf(self.d2(s, k, b, t, sigma, q)) * self.d1(
                s, k, b, t, sigma, q) / sigma
        elif opt_type == 'put':
            return NormalDistribution().pdf(self.d2(s, k, b, t, sigma, q)) * self.d1(
                s, k, b, t, sigma, q) / sigma

    def d_zeta_d_time(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: THE ITM RISK-NEUTRAL PROBABILITY'S SENSITIVITY TO MOVING CLOSER TO MATURITY
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        if opt_type == 'call':
            return self.pdf(self.d2(s, k, b, t, sigma, q)) * (b / (sigma * t ** 0.5) - self.d1(
                s, k, b, t, sigma, q) / (2.0 * t))
        elif opt_type == 'put':
            return -self.pdf(self.d2(s, k, b, t, sigma, q)) * (b / (sigma * t ** 0.5) - self.d1(
                s, k, b, t, sigma, q) / (2.0 * t))

    def risk_neutral_probability_density(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: SECOND ORDER BSM'S FORMULA REGARDING STRIKE
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # return greek
        return NormalDistribution().pdf(self.d2(s, k, b, t, sigma, q)) * np.exp(-r * t) / (
            k * sigma * t ** 0.5)

    def probability_ever_getting_itm(self, s, k, r, t, sigma, q, b, opt_type):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG
        DOCSTRING: THE ITM RISK-NEUTRAL PROBABILITY'S SENSITIVITY TO MOVING CLOSER TO MATURITY
        INPUTS: S (SPOT PRICE), K (STRIKE), R (INTEREST RATE), T (TIME TO MATURITY),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), Q (DIVIDEND YIELD)
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        s, k, r, t, sigma, q, b = self.set_parameters(
            s, k, r, t, sigma, q, b, opt_type)
        # defining greek parameters
        mu = (b - sigma ** 2 / 2.0) / sigma ** 2
        lambda_ = (mu ** 2 + 2 * r / sigma ** 2) ** 0.5
        z = np.log(k / s) / (sigma * t ** 0.5) + lambda_ * sigma * t ** 0.5
        # return greek
        if opt_type == 'call':
            return (k / s) ** (mu + lambda_) * NormalDistribution().cdf(-z) \
                + (k / s) ** (mu - lambda_) * NormalDistribution().cdf(
                -z + 2 * lambda_ * sigma * t ** 0.5)
        elif opt_type == 'put':
            return (k / s) ** (mu + lambda_) * NormalDistribution().cdf(z) \
                + (k / s) ** (mu - lambda_) * NormalDistribution().cdf(
                z - 2 * lambda_ * sigma * t ** 0.5)

    def net_weighted_vega_exposure(self, psi_r, *dicts_opts):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG - PG 119
        DOCSTRING: NET WEIGHTED VEGA EXPOSURE
        INPUTS: PSI_R (VOLATILITY OF REFERENCE VOLATILITY), DICTIONARIES (KEYS - 
            NUMBER OF CONTRACTS WITH VEGA I, T (q_vega), VEGA I, T (vega_t), PSI T (VOLATILITY OF 
            VOLATILITY FOR THE GIVEN SET OF OPTIONS WITH SAME UNDERLYING AND MATURITY) (psi_t), 
            CORRELATION BETWEEN THE VOLATILITY WITH TIME TO MATURITY T AND THE REFERENCE VOLATILITY 
            (corr_t))
        OUTPUTS: FLOAT (VEGA)
        '''
        # initial parameters
        psi_r = self.set_parameters(psi_r)
        # return greek
        return [float(dict_['q_vega']) * float(dict_['vega_t']) * float(dict_['psi_t'])
                * float(dict_['corr_t']) / float(psi_r) for dict_ in dicts_opts].sum()


class IterativeMethods(Greeks):

    def binomial_pricing_model(self, s, k, r, t, n, u, d, opt_type, h_upper=None, h_lower=None):
        '''
        REFERENCES: 
            https://www.youtube.com/watch?v=a3906k9C0fM, 
            https://www.youtube.com/watch?v=WxrRi9lNnqY,
        DOCSTRING: BINOMIAL ASSET PRICING MODEL - UPPER/LOWER BARRIERS FEATURES IMPLEMENTED
        INPUTS: S0 (INITIAL STOCK PRICING MODEL), K (STRIKE PRICE), T (TIME TO MATURITY IN YEARS), 
            R (ANNUAL RISK-FREE RATE), N (NODES), U (UP-FACTOR IN BINOMIAL MODELS), D (DOWN-FACTOR 
            - TO ENSURE RECOMBINING TREE USE 1/U), OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT FROM ARRAY_CP[0] (CALL-PUT PRICING FOR EACH NODE)
        '''
        # initial parameters
        s, k, r, t, n, u, d = self.set_parameters(
            s, k, r, t, n, u, d, opt_type)
        # precomute constants
        dt = t / n
        q = (np.exp(r * dt) - d) / (u - d)
        disc = np.exp(-r * dt)
        # initialise asset prices at maturity - time step n
        array_s_nodes = s * d**(np.arange(n, -1, -1)) * \
            u**(np.arange(0, n + 1, 1))
        # initialise option values at maturity - if intrinsic value is negative, consider zero
        if opt_type == 'call':
            array_cp = np.maximum(array_s_nodes - k, np.zeros(int(n) + 1))
        else:
            array_cp = np.maximum(k - array_s_nodes, np.zeros(int(n) + 1))
        # check s payoff, according to barriers, if values are different from none
        if h_upper != None:
            array_cp[array_s_nodes >= h_upper] = 0
        if h_lower != None:
            array_cp[array_s_nodes <= h_lower] = 0
        # step backwards recursion through tree
        for i in np.arange(int(n), 0, -1):
            array_cp = disc * \
                (q * array_cp[1:i + 1] + (1.0 - q) * array_cp[0:i])
        # returning the no-arbitrage price at node 0
        return array_cp[0]

    def crr_method(self, s, k, r, t, n, sigma, opt_type):
        '''
        REFERENCES: 
            https://www.youtube.com/watch?v=nWslah9tHLk,
            https://quantpy.com.au/binomial-tree-model/binomial-asset-pricing-model-choosing-parameters/
        DOCSTRING: COX, ROSS AND RUBINSTEIN (CRR) METHOD
        INPUTS: SPOT (S), STRIKE (K), RISK-FREE RATE (R), T (TIME TO MATURITY IN YEARS), N (NODES), 
            SIGMA (VOLATILITY OF UNDERLYING ASSET), OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, n, sigma = self.set_parameters(
            s, k, r, t, n, sigma, opt_type)
        # precomute constants
        dt = t / n
        u = np.exp(sigma * np.sqrt(dt))
        d = 1 / u
        q = (np.exp(r * dt) - d) / (u - d)
        disc = np.exp(-r * dt)
        # initialise asset prices at maturity - Time step N
        array_s_nodes = np.zeros(int(n) + 1)
        array_s_nodes[0] = s * d**n
        for j in range(1, int(n) + 1):
            array_s_nodes[j] = array_s_nodes[j - 1] * u / d
        # initialise option values at maturity
        array_cp = np.zeros(int(n) + 1)
        for j in range(0, int(n) + 1):
            #   evaluating maximum value between fair and intrinsic value
            if opt_type == 'call':
                array_cp = np.maximum(array_cp, array_s_nodes - k)
            else:
                array_cp = np.maximum(array_cp, k - array_s_nodes)
        # step backwards through tree
        for i in np.arange(int(n), 0, -1):
            for j in range(0, i):
                array_cp[j] = disc * \
                    (q * array_cp[j + 1] + (1 - q) * array_cp[j])
        # returning the no-arbitrage price at node 0
        return array_cp[0]

    def jr_method(self, s, k, r, t, n, sigma, opt_type):
        '''
        REFERENCES: 
            https://www.youtube.com/watch?v=nWslah9tHLk,
            https://quantpy.com.au/binomial-tree-model/binomial-asset-pricing-model-choosing-parameters/
        DOCSTRING: JARROW AND RUDD (JR) METHOD
        INPUTS: SPOT (S), STRIKE (K), RISK-FREE RATE (R), T (TIME TO MATURITY IN YEARS), N (NODES), 
            SIGMA (VOLATILITY OF UNDERLYING ASSET), OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, n, sigma = self.set_parameters(
            s, k, r, t, n, sigma, opt_type)
        # precomute constants
        dt = t / n
        nu = r - 0.5 * sigma**2
        u = np.exp(nu * dt + sigma * np.sqrt(dt))
        d = np.exp(nu * dt - sigma * np.sqrt(dt))
        q = 0.5
        disc = np.exp(-r * dt)
        # initialise asset prices at maturity - Time step N
        array_s_nodes = np.zeros(int(n) + 1)
        array_s_nodes[0] = s * d**n
        for j in range(1, int(n) + 1):
            array_s_nodes[j] = array_s_nodes[j - 1] * u / d
        # initialise option values at maturity
        array_cp = np.zeros(int(n) + 1)
        for j in range(0, int(n) + 1):
            #   evaluating maximum value between fair and intrinsic value
            if opt_type == 'call':
                array_cp = np.maximum(array_cp, array_s_nodes - k)
            else:
                array_cp = np.maximum(array_cp, k - array_s_nodes)
        # step backwards through tree
        for i in np.arange(int(n), 0, -1):
            for j in range(0, i):
                array_cp[j] = disc * \
                    (q * array_cp[j + 1] + (1 - q) * array_cp[j])
        # returning the no-arbitrage price at node 0
        return array_cp[0]

    def eqp_method(self, s, k, r, t, n, sigma, opt_type):
        '''
        REFERENCES: 
            https://www.youtube.com/watch?v=nWslah9tHLk,
            https://quantpy.com.au/binomial-tree-model/binomial-asset-pricing-model-choosing-parameters/
        DOCSTRING: EQUAL PROBABILITIES (EQP) METHOD
        INPUTS: SPOT (S), STRIKE (K), RISK-FREE RATE (R), T (TIME TO MATURITY IN YEARS), N (NODES), 
            SIGMA (VOLATILITY OF UNDERLYING ASSET), OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, n, sigma = self.set_parameters(
            s, k, r, t, n, sigma, opt_type)
        # precomute constants
        dt = t / n
        nu = r - 0.5 * sigma**2
        dxu = 0.5 * nu * dt + 0.5 * \
            np.sqrt(4 * sigma**2 * dt - 3 * nu**2 * dt**2)
        dxd = 1.5 * nu * dt - 0.5 * \
            np.sqrt(4 * sigma**2 * dt - 3 * nu**2 * dt**2)
        pu = 0.5
        pd = 1 - pu
        disc = np.exp(-r * dt)
        # initialise asset prices at maturity - Time step N
        array_s_nodes = np.zeros(int(n) + 1)
        array_s_nodes[0] = s * np.exp(n * dxd)
        for j in range(1, int(n) + 1):
            array_s_nodes[j] = array_s_nodes[j - 1] * np.exp(dxu - dxd)
        # initialise option values at maturity
        array_cp = np.zeros(int(n) + 1)
        for j in range(0, int(n) + 1):
            #   evaluating maximum value between fair and intrinsic value
            if opt_type == 'call':
                array_cp = np.maximum(array_cp, array_s_nodes - k)
            else:
                array_cp = np.maximum(array_cp, k - array_s_nodes)
        # step backwards through tree
        for i in np.arange(int(n), 0, -1):
            for j in range(0, i):
                array_cp[j] = disc * \
                    (pu * array_cp[j + 1] + pd * array_cp[j])
        # returning the no-arbitrage price at node 0
        return array_cp[0]

    def trg_method(self, s, k, r, t, n, sigma, opt_type):
        '''
        REFERENCES: 
            https://www.youtube.com/watch?v=nWslah9tHLk,
            https://quantpy.com.au/binomial-tree-model/binomial-asset-pricing-model-choosing-parameters/
        DOCSTRING: TRIGEORGIS (TRG) METHOD
        INPUTS: SPOT (S), STRIKE (K), RISK-FREE RATE (R), T (TIME TO MATURITY IN YEARS), N (NODES), 
            SIGMA (VOLATILITY OF UNDERLYING ASSET), OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT
        '''
        # initial parameters
        s, k, r, t, n, sigma = self.set_parameters(
            s, k, r, t, n, sigma, opt_type)
        # precomute constants
        dt = t / n
        nu = r - 0.5 * sigma**2
        dxu = np.sqrt(sigma**2 * dt + nu**2 * dt**2)
        dxd = -dxu
        pu = 0.5 + 0.5 * nu * dt / dxu
        pd = 1 - pu
        disc = np.exp(-r * dt)
        # initialise asset prices at maturity - Time step N
        array_s_nodes = np.zeros(int(n) + 1)
        array_s_nodes[0] = s * np.exp(n * dxd)
        for j in range(1, int(n) + 1):
            array_s_nodes[j] = array_s_nodes[j - 1] * np.exp(dxu - dxd)
        # initialise option values at maturity
        array_cp = np.zeros(int(n) + 1)
        for j in range(0, int(n) + 1):
            #   evaluating maximum value between fair and intrinsic value
            if opt_type == 'call':
                array_cp = np.maximum(array_cp, array_s_nodes - k)
            else:
                array_cp = np.maximum(array_cp, k - array_s_nodes)
        # step backwards through tree
        for i in np.arange(int(n), 0, -1):
            for j in range(0, i):
                array_cp[j] = disc * \
                    (pu * array_cp[j + 1] + pd * array_cp[j])
        # returning the no-arbitrage price at node 0
        return array_cp[0]


class EuropeanOptions(IterativeMethods):

    @lru_cache()
    def implied_volatility(self, s, k, r, t, sigma, q, b, cp0, opt_type, 
                           method='fsolve',
                           tolerance=1E-3, epsilon=1, max_iter=1000, orig_vol=0.5,
                           list_bounds=[(0, 2)]):
        '''
        REFERENCES: https://www.youtube.com/watch?v=Jpy3iCsijIU, 
            https://www.option-price.com/documentation.php#impliedvolatility
        DOCSTRING: CALCULATING THE IMPLIED VOLATILITY FOR A GIVEN
        INPUTS: OPTION CALL OR PUT, TOLERANCE, EPSILON, MAX ITERATIONS, ORIGINAL VOL
        OUTPUTS: IMPLIED VOLATILITY AND MAX ITERATION HITTED BOOLEAN, ALL ENCAPSULED IN A TUPLE
        '''
        # initial parameters
        s, k, r, t, sigma, q, b, cp0 = self.set_parameters(
            s, k, r, t, sigma, q, b, cp0, opt_type)
        count = 0
        if method == 'newton_raphson':
            # iterating until the error is meaningless
            while epsilon > tolerance:
                # passing parameters
                count += 1
                flag_max_iter_hitten = False
                if count == 1:
                    imp_vol = orig_vol
                # preventing infinite loop
                if count >= max_iter:
                    flag_max_iter_hitten = True
                    break
                # calculating the difference betwwen call prize, by the implied vol and call price
                dif_calc_market = self.general_opt_price(s, k, r, t, imp_vol, q, b, opt_type) \
                    - cp0
                # newthon-hampson-model to check whether the zero of the function has been spoted
                #   working with a tolerance to assume the zero of the function has been found
                # if self.vega(s, k, r, t, imp_vol, q, b) != 0:
                try:
                    imp_vol = -dif_calc_market / \
                        self.vega(s, k, r, t, imp_vol, q, b) + imp_vol
                except RuntimeWarning:
                    return imp_vol, flag_max_iter_hitten
                # else:
                #     raise Exception("Vega musn't be zero")
                epsilon = abs((imp_vol - orig_vol) / imp_vol)
            # returning implied volatility and maximum iterations hitten
            return imp_vol, flag_max_iter_hitten
        elif method == 'bisection':
            float_high = 5
            float_low = 0
            while (float_high - float_low) > epsilon:
                if self.general_opt_price(s, k, r, t, float(float_high + float_low) / 2.0, q, 
                                          b, opt_type) > cp0:
                    float_high = float(float_high + float_low) / 2.0
                else: 
                    float_low = float(float_high + float_low) / 2.0
            return (float_high + float_low) / 2
        elif method == 'fsolve':
            # defining a non-linear function as the difference of the theoratical price (Black
            #   & Scholes) and the fair price
            def func_non_linear(sigma): return np.power(
                self.general_opt_price(s, k, r, t, sigma, q, b, opt_type) - cp0, 2)
            # print(func_non_linear(orig_vol))
            # returning the least sigma for the cost function
            return fsolve(func_non_linear, orig_vol)
        elif method == 'scipy_optimize_minimize':
            # defining a non-linear function as the difference of the theoratical price (Black
            #   & Scholes) and the fair price
            def func_non_linear(sigma): return np.power(
                self.general_opt_price(s, k, r, t, sigma, q, b, opt_type) - cp0, 2)
            # print(func_non_linear(orig_vol))
            # returning the least sigma for the cost function
            return minimize(func_non_linear, orig_vol, method='CG')
        elif method == 'differential_evolution':
            # defining a non-linear function as the difference of the theoratical price (Black
            #   & Scholes) and the fair price
            def func_non_linear(sigma): return np.power(
                self.general_opt_price(s, k, r, t, sigma, q, b, opt_type) - cp0, 2)
            # print(func_non_linear(orig_vol))
            # returning the least sigma for the cost function
            return NonLinearEquations().differential_evolution(func_non_linear, list_bounds)
        else:
            raise Exception('Method to return the root of the non-linear equation is not '
                            + 'recognized, please revisit the parameter')

    def moneyness(self, s, k, r, t, sigma, q):
        '''
        REFERENCES: MERCADO DE OPÇÕES, CONCEITOS E ESTRATÉGIAS / AUTOR: LUIZ MAURÍCIO DA SILVA /
            PGS. 74, 75, 76, 77, 78
        DOCSTRING: MEASURES WHETER THE OPTION WILL BE EXERCISED OR NOT TRANSLATED IN A PERCENTUAL
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY), R (INTEREST RATE) AND
            SIGMA (VOLATILITY OF UNDERLYING ASSET)
        OUTPUTS: PERCENTAGE
        '''
        # initial parameters
        s, k, r, t, sigma, q = self.set_parameters(s, k, r, t, sigma, q)
        # returning moneyness
        return (self.d1(s, k, r, t, sigma, q)
                + self.d2(s, k, r, t, sigma, q)) / 2

    def iaotm(self, s, k, r, t, sigma, opt_type, pct_moneyness_atm=0.05):
        '''
        DOCSTRING: ITM / ATM / OTM - OPTIONS PREDICT OF EXERCISING
        INPUTS: S (SPOT PRICE), K (STRIKE), T (TIME TO MATURITY), R (INTEREST RATE),
            SIGMA (VOLATILITY OF UNDERLYING ASSET), OPTION TYPE AND PERCENTAGE OF ATM
            (STANDARD VALUE OF 5%)
        OUTPUTS: ITM/ATM/OTM
        '''
        # initial parameters
        s, k, r, t, sigma, q = self.set_parameters(
            s, k, r, t, sigma, q, opt_type)
        # determining iaotm
        if abs(self.moneyness(s, k, r, t, sigma)) < pct_moneyness_atm:
            return 'ATM'
        elif (self.moneyness(s, k, r, t, sigma) < pct_moneyness_atm and
              opt_type == 'call') or (self.moneyness(s, k, r, t, sigma) >
                                       pct_moneyness_atm and opt_type == 'put'):
            return 'OTM'
        elif (self.moneyness(s, k, r, t, sigma) > pct_moneyness_atm and
              opt_type == 'call') or (self.moneyness(s, k, r, t, sigma) <
                                       pct_moneyness_atm and opt_type == 'put'):
            return 'ITM'
        else:
            raise Exception(
                'Please revisit your inputs, request did not return appropriate values')


# print(BlackScholesMerton(50, 45, 0.02, 80 / 365, 0.3).call_price())
# # output: 6.021277654922962
# print(BlackScholesMerton(100, 105, 0.01, 30 / 365, 0.38).call_price())
# print(BlackScholesMerton(100, 105, 0.01, 30 / 365, 0.15).call_price())

# # option parameters
# s = 100.0
# k = 105.0
# t = 30.0 / 365.0
# r = 0.01
# q = 0
# sigma = 0
# c0 = 2.30
# opt_type = 'call'

# # result
# print(EuropeanOptions().implied_volatility(opt_type))
# # output: (0.3688563249135555, True)

# # result
# print(EuropeanOptions(100.0, 105.0, 0.01, 30.0 / 365.0,
#                      0, 0, 2.30, 365).implied_volatility('call'))
# # output: (0.3688563249135555, True)

# print(EuropeanOptions().implied_volatility(194.11, 210.0, 0.01, 38 / 365,
#                                           0, 0, 1.50, 'call'))
# # output: (0.25273873689374843, True)

# print(EuropeanOptions().call_price(50, 45, 0.02, 80 / 365, 0.3))
# # output: 6.021277654922962

# print(EuropeanOptions().delta_call(100, 105, 0.01, 30 / 365, 0.31, 0))
# # output: 0.3101960860984582

# print(EuropeanOptions().delta_put(100, 105, 0.01, 30 / 365, 0.31, 0))
# # output: -0.6898039139015417

# print(EuropeanOptions().gamma(100, 105, 0.01, 30 / 365, 0.31, 0))
# # output: 0.03970674807221808

# print(EuropeanOptions().theta_call(819.42, 1020, 0.01, 42 / 365, 0.6966, 0))
# # output: -0.6700017174199313

# print(EuropeanOptions().theta_put(819.42, 1020, 0.01, 42 / 365, 0.6966, 0))
# # output: -0.6420886495736411

# print(EuropeanOptions().delta_call(100, 105, 0.01, 30 / 365, 0.31, 0))
# # output: 0.3101960860984582

# print(EuropeanOptions().delta_put(100, 105, 0.01, 30 / 365, 0.31, 0))
# # output: -0.6898039139015417

# print(EuropeanOptions().gamma(100, 105, 0.01, 30 / 365, 0.31, 0))
# # output: 0.03970674807221808

# print(EuropeanOptions().implied_volatility(
#     100.0, 105.0, 0.01, 30.0 / 365.0, 0, 2.30, 'call'))
# # output: (0.3688563249135555, True)

# print(EuropeanOptions().basic_opt_price(50, 45, 0.02, 80 / 365, 0.3, 0, 'call'))
# # output: 6.021277654922962

# print(EuropeanOptions().basic_opt_price(50, 45, 0.02, 80 / 365, 0.3, 0, 'put'))
# # output: 0.8244491011813224

# print(EuropeanOptions().implied_volatility(194.11, 210.0, 0.01, 38 / 365,
#                                           0, 1.50, 'call'))
# # output: (0.25273873689374843, True)

# print(EuropeanOptions().implied_volatility(100.0, 95.0, 0.01, 30 / 365,
#                                           0, 2.30, 'put'))
# # output: (0.3931007096283138, True)

# print(EuropeanOptions().implied_volatility(100.0, 95.0, 0.01, 30 / 365,
#                                           0, 2.30, 'put'))

# print(EuropeanOptions().implied_volatility(
#     14.50, 15, 0.0225, 10 / 365, 0.0861, 0.25, 'call'))
# # output:(0.47974617961730276, True)

# print(EuropeanOptions().implied_volatility(
#     17.65, 15, 0.0225, 467 / 252, 0, 7.61, 'call'))
# # output:(0.6983823635729837, True)


# print(EuropeanOptions().implied_volatility(
#     11.33, 14.50, 0.0225, 3 / 365, 0.09076773878571428, 0.01, 'call',
#     orig_vol=0.812424814792))


# def newton_vol_call_div(s, k, t, c, r, q, sigma):

#     fx = s * np.exp(-q * t) * si.norm.cdf(EuropeanOptions().d1(s, k, r, t, sigma), 0.0, 1.0) - \
#         k * np.exp(-r * t) * si.norm.cdf(EuropeanOptions().d2(s,
#                                                              k, r, t, sigma), 0.0, 1.0) - c

#     vega = (1 / np.sqrt(2 * np.pi)) * s * np.exp(-q * t) * np.sqrt(t) * \
#         np.exp(
#             (-si.norm.cdf(EuropeanOptions().d1(s, k, r, t, sigma), 0.0, 1.0) ** 2) * 0.5)

#     tolerance = 0.000001
#     x0 = sigma
#     xnew = x0
#     xold = x0 - 1

#     while abs(xnew - xold) > tolerance:

#         xold = xnew
#         xnew = (xnew - fx - c) / vega

#         return abs(xnew)


# print(newton_vol_call_div(11.33, 14.50, 3 / 365,
#                           0.08, 0.0225, 0.09076773878571428, 0.812424814792))

# max(0.041276451234, '[]')

# print(EuropeanOptions().delta(5127.7930, 5300.0, 0.02, 60 / 365, 0.3, 0, 'call'))

# s = 24.87
# k = 40.25
# r = 0.0775
# t = 0.05555555555555555
# sigma = 1.9251923611193833

# print(EuropeanOptions().d1(s, k, r, t, sigma))
# # output: -0.8246155162915018



### MODULE TO DEAL WITH PDF, FROM EXTRACTION TO HANDLING DATA ###

import tabula
import textwrap
from PyPDF2 import PdfFileReader
from fpdf import FPDF
from base64 import b64encode
from stpstone.handling_data.folders import DirFilesManagement


class PDFHandler:

    def fetch(self, complete_pdf_name, method_open='rb', str_return_num_pages_text='num_pages'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # import to memory pdf object
        pdf_file_object = open(complete_pdf_name, method_open)
        # pdf reader object
        pdf_reader = PdfFileReader(pdf_file_object)
        # return number of pages and text
        if str_return_num_pages_text == 'num_pages':
            return pdf_reader.numPages
        elif str_return_num_pages_text == 'text_pages':
            return [pdf_reader.getPage(num_page).extractText() for num_page in range(
                1, pdf_reader.numPages)]

    def extract_tables(self, complete_pdf_name, str_num_pages):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # returning table data withn pages of pdf file
        return tabula.read_pdf(complete_pdf_name, pages=str_num_pages)

    def pdf_to_base64(self, filename):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        with open(filename, 'rb') as pdf_file:
            return b64encode(pdf_file.read()).decode()

    def text_to_pdf(text, filename, a4_width_mm=210, pt_to_mm=0.35, fontsize_pt=10,
                    margin_bottom_mm=10,
                    character_width=7, orientation='P', unit='mm', format='A4', font_family='Courier',
                    output_file='F'):
        '''
        REFERENCES: https://stackoverflow.com/questions/10112244/convert-plain-text-to-pdf-in-python
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        fontsize_mm = fontsize_pt * pt_to_mm
        character_width_mm = character_width * pt_to_mm
        width_text = a4_width_mm / character_width_mm
        # set pdf layout
        pdf = FPDF(orientation=orientation, unit=unit, format=format)
        pdf.set_auto_page_break(True, margin=margin_bottom_mm)
        pdf.add_page()
        pdf.set_font(family=font_family, size=fontsize_pt)
        splitted = text.split('\n')
        # loop through lines of text
        for line in splitted:
            lines = textwrap.wrap(line, width_text)
            if len(lines) == 0:
                pdf.ln()
            for wrap in lines:
                pdf.cell(0, fontsize_mm, wrap, ln=1)
        pdf.output(filename, output_file)
        # checking wheter the file was created
        return DirFilesManagement().object_exists(filename)



### AWS S3 CLIENT ###

# pypi.org libs
import boto3
import sys
from botocore.exceptions import NoCredentialsError, ClientError, EndpointConnectionError
from keyring import get_password
from logging import Logger
from io import BytesIO
from typing import Optional, List, Dict, Any
# local libs
from stpstone.loggs.create_logs import CreateLog


class S3Client:

    def __init__(self, str_default_region:str='us-west-1', logger:Optional[Logger]=None, 
                 bl_debug_mode:bool=False) -> None:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUT:S
        '''
        self._envs = {
            'aws_access_key_id': get_password('AWS_ACCESS', 'KEY_ID'),
            'aws_secret_access_key': get_password('AWS_ACCESS', 'PASSWORD'),
            'region_name': get_password('AWS_ACCESS', 'REGION_NAME') \
                if get_password('AWS_ACCESS', 'REGION_NAME') is not None \
                else str_default_region,
            's3_bucket': get_password('S3_BUCKET_NAME'),
            'datalake': get_password('DELTA_LAKE_S3_PATH')
        }
        for var in self._envs:
            if self._envs[var] is None:
                print(f'A variável de ambiente {var} não está definida.')
                sys.exit(1)
        self.s3 = boto3.client(
            's3',
            aws_access_key_id=self._envs['aws_access_key_id'],
            aws_secret_access_key=self._envs['aws_secret_access_key'],
            region_name=self._envs['region_name']
        )
        self.logger = logger
        self.bl_debug_mode = bl_debug_mode

    def handle_error(self, e:Exception, action:Optional[str]=None, s3_key:Optional[str]=None) \
        -> None:
        '''
        DOCSTRING: GENERIC ERROR HANDLER
        INPUTS: EXCEPTION OBJECT, ACTION (OPTIONAL), S3 KEY (OPTIONAL)
        OUTPUTS: LOGGER/PRINT
        '''
        if isinstance(e, NoCredentialsError):
            message = 'Credentials not found, please reconfigure your AWS credentials properly.'
        elif isinstance(e, ClientError):
            if e.response['Error']['Code'] == "NoSuchKey":
                message = f'The file {s3_key} was not found in the bucket {self._envs["s3_bucket"]}.'
            else:
                message = f'AWS Client Error: {e}'
        elif isinstance(e, EndpointConnectionError):
            message = 'Failed to connect to the AWS endpoint. Please check your internet connection.'
        elif isinstance(e, FileNotFoundError):
            message = f'The file {s3_key} was not found locally.'
        else:
            message = f'An unexpected error occurred during {action or "an operation"}: {e}'
        # logging or printing, in case of debug mode
        if self.logger:
            CreateLog().critical(self.logger, message)
        if self.bl_debug_mode:
            print(message)

    def upload_file(self, data:BytesIO, s3_key:str) -> None:
        '''
        DOCSTRING: UPLOAD FILES TO AWS CLOUD
        INPUTS: DATA, S3 KEY
        OUTPUTS:
        '''
        try:
            self.s3.put_object(
                Body=data.getvalue(),
                Bucket=self._envs['s3_bucket'],
                Key=s3_key
            )
        except Exception as e:
            self.handle_error(e, action='uploading a file', s3_key=s3_key)

    def download_file(self, s3_key:str) -> Optional[Dict[str, Any]]:
        '''
        DOCSTRING: DOWNLOAD FILE FROM S3
        INPUTS: S3 KEY
        OUTPUTS: DOWNLOADED FILE
        '''
        try:
            file = self.s3.get_object(Bucket=self._envs['s3_bucket'], Key=s3_key)
            print(f'Successfully downloaded {s3_key}')
            return file
        except Exception as e:
            self.handle_error(e, action='downloading a file', s3_key=s3_key)

    def list_object(self, prefix:str) -> Optional[List[Dict[str, Any]]]:
        '''
        DOCSTRING: LIST OBJECTS IN S3
        INPUTS: PREFIX
        OUTPUTS: LIST OF OBJECTS
        '''
        try:
            return self.s3.list_objects(Bucket=self._envs['s3_bucket'], Prefix=prefix)['Contents']
        except Exception as e:
            self.handle_error(e, action='listing objects')



### GLOBAL CONSTANTS ###

# pypi.org libs
import os
# local libs
from stpstone.opening_config.setup import reading_yaml


# base path
path_base = os.path.dirname(os.path.realpath(__file__))
# slots of memory to each yaml
YAML_ANBIMA = reading_yaml(os.path.join(path_base, 'anbima.yaml'))
YAML_B3 = reading_yaml(os.path.join(path_base, 'b3.yaml'))
YAML_BR_TRS = reading_yaml(os.path.join(path_base, 'br_treasury.yaml'))
YAML_CD = reading_yaml(os.path.join(path_base, 'comdinheiro.yaml'))
YAML_GEN = reading_yaml(os.path.join(path_base, 'generic.yaml'))
YAML_WGBD = reading_yaml(os.path.join(path_base, 'world_gov_bonds.yaml'))
YAML_GLB_RT = reading_yaml(os.path.join(path_base, 'global_rates.yaml'))
YAML_USA_MACRO = reading_yaml(os.path.join(path_base, 'usa_macro.yaml'))
YAML_BR_MACRO = reading_yaml(os.path.join(path_base, 'br_macro.yaml'))
YAML_MICROSOFT_APPS = reading_yaml(os.path.join(path_base, 'microsoft_apps.yaml'))
YAML_LLMS = reading_yaml(os.path.join(path_base, 'llms.yaml'))
YAML_SESSION = reading_yaml(os.path.join(path_base, 'session.yaml'))


### CAPITAL RISK ASSESSMENT ###

class CR1Calculator:
    '''
    DOCSTRING: A CLASS TO CALCULATE CAPITAL REQUIREMENT 1 (CR1) FOR FINANCIAL INSTRUMENTS
    INPUTS: EAD, PD, LGD
    OUTPUTS: K, CR1
    '''
    def __init__(self, float_ead: float, float_pd: float, float_lgd: float):
        '''
        DOCSTRING: INITIALIZE THE CALCULATION WITH BASIC PARAMETERS
        INPUTS:
            - EXPOSURE AT DEFAULT (EAD) IN MONETARY UNITS
            - PROBABILITY OF DEFAULT (PD) AS A FRACTION
            - LOSS GIVEN DEFAULT (LGD) AS A FRACTION
        '''
        self.float_ead = float_ead
        self.float_pd = float_pd
        self.float_lgd = float_lgd

    @property
    def calculate_k(self) -> float:
        '''
        DOCSTRING: CALCULATE THE RISK FACTOR (K) BASED ON THE PARAMETERS
        INPUTS: -
        OUTPUTS: FLOAT
        '''
        # Simplified calculation for K, regulatory adjustments (R) can be added accordingly
        return self.float_lgd * self.float_pd

    @property
    def calculate_cr1(self) -> float:
        '''
        DOCSTRING: CALCULATE THE CAPITAL REQUIREMENT 1 (CR1)
        INPUTS: -
        OUTPUTS: VALUE OF CR1 IN MONETARY UNITS.
        '''
        return 12.5 * self.calculate_k * self.float_ead

    @property
    def summary(self) -> dict:
        '''
        DOCSTRING: RETURN A SUMMARY OF THE PARAMETERS AND CALCULATIONS.
        INPUTS: - 
        OUTPUTS: DICTIONARY CONTAINING EAD, PD, LGD, K, AND CR1.
        '''
        return {
            'EAD': self.float_ead,
            'PD': self.float_pd,
            'LGD': self.float_lgd,
            'K': self.calculate_k,
            'CR1': self.calculate_cr1
        }